#include "stdafx.h"
#include "math.h"
#include <afxcoll.h>
#include "BINModel.h"
#include "TexListDlg.h"
#include "C:\dev\code\generic classes\optools.h"

#include "gl\gl.h"
#include "gl\glu.h"
#include "gl\glaux.h"

CBINModel::CBINModel() {
	bIsInitialized = FALSE;

	cmapTexPtrs.RemoveAll();
	cmapBmpTexPtrs.RemoveAll();
	cmapTexSizes.RemoveAll();

	InitFaceTypes();
}

CBINModel::~CBINModel() {
	int i;

	// clear BIN model
	NewBIN();

	// clean tex_pointers, tex_names Array
	void *pBuf;
	for (i=0; i<csaBINTextures.GetSize(); i++) {
		if (cmapTexPtrs.Lookup(csaBINTextures[i], pBuf)) free(pBuf);
		if (cmapBmpTexPtrs.Lookup(csaBINTextures[i], pBuf)) free(pBuf);
	}
	cmapTexPtrs.RemoveAll();
	cmapBmpTexPtrs.RemoveAll();
	cmapTexSizes.RemoveAll();

	// texture animations
	sTexAnim *psTemp;
	for (i=0; i<cpaTexAnims.GetSize(); i++) {
		psTemp = (sTexAnim *)cpaTexAnims.GetAt(i);
		psTemp->csName.Empty();
		psTemp->csaTextures.RemoveAll();
		delete psTemp;
	}
	cpaTexAnims.RemoveAll();
}

int CBINModel::ReadBIN(CString filename, CString csARTDir)
{
	if (!bIsInitialized) return 0;

	int i;
	int iOldFaceNum = GetNumberOfFaces();
	int iOldVertNum = GetNumberOfVertices();
	int num, coord;
	CString csMsg;

	// Open the BIN file
	CFile bin;
	name = filename;
	
	if (!bin.Open(filename, CFile::modeRead, NULL)) {
		csMsg.Format("CBINModel::ReadBIN() - File not found:\n%s", filename);
		AfxMessageBox(csMsg);
		return -1;
	}
	bin.Read(&num, 4);
	if (num == 0x20) {
		AfxMessageBox("This is an animation control BIN that\ndoes not contain any BIN model data.");
		bin.Close();
		return -1;
	} else if (num != 0x14) {
		csMsg.Format("Unknown BIN type 0x%x.\nBINedit may crash when you attempt to load it.\nContinue loading?", num);
		if (AfxMessageBox(csMsg, MB_YESNO) != IDYES)	{
			bin.Close();
			return -1;
		}
	}
	bin.Read(&liHellMagic, 4);	// 0x0001 0000 for non Hellbender

	// read vertices
	bin.Seek(0x10, CFile::begin);
	bin.Read(&num, 4);
	TRACE("CBINModel::ReadBIN(): Number Of Vertices = %d\n", num);
	for (i=0; i<3*num; i++) {
		bin.Read(&coord, 4);
		vertices.Add((UINT)coord);
	}

	// texture block
	char buf[100];
	CString texture;
	int aux, tex_size;
	DWORD dwAux;
	face *curface;

	while (bin.Read(&num, 4)) {
		if (num == 0x03) {	// something MTM2 specific, I think (Vertex Normals)
			TRACE("  Data Block 0x03 - ignored on read!\n");
			bin.Read(&num, 4);	// dummy
			bin.Read(&num, 4);	// number of vertices?
			for (i=0; i<num; i++) {
				bin.Read(&aux, 4);
				bin.Read(&aux, 4);
				bin.Read(&aux, 4);
			}

		} else if (num == 0x17) {	// helbender, just ignore 8 bytes on read...
			TRACE("  Data Block 0x17 - ignore\n");
			bin.Read(&num, 4);
			TRACE(" %d (%x)\n", num, num);
			bin.Read(&num, 4);
			TRACE(" %d (%x)\n", num, num);
		} else if (num == 0x0D) {
			bin.Read(buf, 4);
			bin.Read(buf, 12);
			buf[12]=0;
			TRACE("  Texture: %s\n", buf);
			texture = CString(buf);
			texture.MakeUpper();
			bin.Read(buf, 4);
			aux = LoadTexture(texture, csARTDir);
			if (aux != -1) tex_size = aux;
			else {
				bin.Close();
				return -1;
			}
		} else if (num == 0x1D) {
			WORD wNum;
			WORD wDelay;
			CStringArray csaTexs;
			
			bin.Read(buf, 4);
			bin.Read(&wNum, 4);		// number of textures
			bin.Read(buf, 4);
			bin.Read(&wDelay, 4);	// animation delay
			bin.Read(buf, 8);
			
			for (int i=0; i<wNum; i++) {
				bin.Read(buf, 32);
				TRACE("Anim Texture %d: %s\n", i, buf);
				texture = buf;
				texture.MakeUpper();
				aux = LoadTexture(texture, csARTDir);
				if (aux != -1) tex_size = aux;
				else {
					bin.Close();
					return -1;
				}
				csaTexs.Add(texture);
			}

			texture = "Loaded"+CString(itoa(wAnimCnt++,buf,10));
			TRACE("Animation Name: %s\n", texture);
			AddTexAnim(texture, wDelay, csaTexs, csARTDir);
			texture = "ANIM:"+texture;
			
		} else if (num == 0x18 || num == 0x11 || num == 0x29 || num == 0x33 || num == 0xE) {
			TRACE("  Face Type 0x%X\n", num);
			curface = new face;	
			curface->texture = texture;
			curface->type = num;
			curface->tex_size = tex_size;

			bin.Read(&num, 4);	// number of vertices
			//TRACE("%d sided polygon type %d\n", num, curface->type);
			if (num > 4) {
				AfxMessageBox("Unsupported number of vertices!");
				bin.Close();
				return 0;
			}
			bin.Read(&(curface->normal[0]), 4);
			bin.Read(&(curface->normal[1]), 4);
			bin.Read(&(curface->normal[2]), 4);
			bin.Read(&(curface->unknown), 4);
			for (int i=0; i<num; i++) {
				bin.Read(&aux, 4);
				curface->points.Add(aux+iOldVertNum);
				bin.Read(&aux, 4);
				curface->tex_x.Add(aux);
				bin.Read(&aux, 4);
				curface->tex_y.Add(aux);
			}
			faces.Add(curface);
		} else if (num == 0x19) {
			TRACE("  Face Type 0x19\n", num);
			curface = new face;	
			curface->texture = texture;
			curface->type = num;
			curface->tex_size = tex_size;

			bin.Read(&num, 4);	// number of vertices
			if (num > 4) {
				AfxMessageBox("Unsupported number of vertices!");
				bin.Close();
				return 0;
			}
			bin.Read(&(curface->normal[0]), 4);
			bin.Read(&(curface->normal[1]), 4);
			bin.Read(&(curface->normal[2]), 4);
			bin.Read(&(curface->unknown), 4);
			for (int i=0; i<num; i++) {
				bin.Read(&aux, 4);
				curface->points.Add(aux+iOldVertNum);
				curface->tex_x.Add(0);
				curface->tex_y.Add(0);
			}
			faces.Add(curface);
		} else if (num == 0x00 || num == 0x0A) {
			TRACE("  Face Type 0x%X - nothing read\n", num);
		} else {
			TRACE("Unknown face type %d (0x%X)\n", num, num);
			csMsg.Format("The BIN model uses an unsupported face type!\nType: 0x%x at %ld (0x%lx)", num, bin.GetPosition(),  bin.GetPosition());
			AfxMessageBox(csMsg);
			bin.Close();
			return -1;
		}
	}

	bin.Close();

	// Read .INFO file, if it exists
	ReadInfoFile(name, iOldVertNum, iOldFaceNum);

	faces_tag.SetSize(GetNumberOfFaces());
	vertices_tag.SetSize(GetNumberOfVertices());
	for (i=0; i<faces_tag.GetSize(); i++) faces_tag[i] = 0;
	for (i=0; i<vertices_tag.GetSize(); i++) vertices_tag[i] = 0;

	return 1;
}


void CBINModel::DrawBIN(int FaceType, BOOL bAxes, BOOL bVerts)
#define FT_VERTEX	(0)
#define FT_WIRE		(1)
#define FT_FILLED	(2)
#define FT_TEXTURE	(3)
{
/*	//Display Lists didn't prove to be faster
	static GLuint iList = 1;

	if (!bRebuildDisplayList) {
		static BYTE baLists[1];
		baLists[0] = 1;
		glCallLists(1, GL_BYTE, baLists);
		return;
	}

	glNewList(iList, GL_COMPILE_AND_EXECUTE); */

	static int i;
	static double x,y,z, tx, ty;
	static face *curface;
	CString csOldtex = "";
	BOOL oldTagged = FALSE;
	BOOL oldTriangle = TRUE;

	glColor3f (1.0f, 1.0f, 1.0f);
	
	// draw all faces
	for (i=0; i<GetNumberOfFaces(); i++) {

		if (i == current_face) continue;	// except the "current" face

		curface = (face *)faces[i];

		// Texture benutzen...
		if (FaceType == FT_TEXTURE) {
			if (csOldtex != curface->texture) {
				csOldtex = curface->texture;

				if (curface->texture.Left(5)=="ANIM:") {
					TRACE("Animated Texture: %s\n",csOldtex);
					TRACE("TexSize : %d\n",curface->tex_size);
					csOldtex=GetFirstAnimTexture(csOldtex.Right(csOldtex.GetLength()-5));
				}

				glTexImage2D(GL_TEXTURE_2D, 0, 3, curface->tex_size, 
				curface->tex_size, 0, GL_RGB, GL_UNSIGNED_BYTE, 
				GetTexturePointer(csOldtex));
			}
		}


		glBegin(GL_TRIANGLES);
		oldTriangle = TRUE;

		if (FaceType >= FT_FILLED) {

			// tagged faces = green
			if (oldTagged != faces_tag[i]) {
				oldTagged = faces_tag[i];
				if (!oldTagged) glColor3f (1.0f, 1.0f, 1.0f);
				else glColor3f (0.0f, 1.0f, 0.0f);

			}
						
			if (oldTriangle != (curface->points.GetSize()==3)) {
				glEnd();
				oldTriangle = (curface->points.GetSize()==3);
			    if (oldTriangle) glBegin(GL_TRIANGLES);
				else  glBegin(GL_QUADS);
			}

			for (int j=0; j<curface->points.GetSize(); j++) {
				x = (double)((int)vertices[3*curface->points[j]]);
				y = (double)((int)vertices[3*curface->points[j]+1]);
				z = -(double)((int)vertices[3*curface->points[j]+2]);
				x /= 1000;
				y /= 1000;
				z /= 1000;
				tx = (double)(curface->tex_x[j])/(double)(0xFF0000);
				ty = (double)(curface->tex_y[j])/(double)(0xFF0000);

				x *= zoom; y *= zoom; z *= zoom;
				
				// Looks better w/o normal vector
				if (FaceType == FT_FILLED)  glNormal3i(curface->normal[0],curface->normal[1],curface->normal[2]);
				if (FaceType == FT_TEXTURE) glTexCoord2f(tx, ty);
				glVertex3d(x, y, z);
			}
		}	// if >= FT_FILLED
		glEnd();

		if (FaceType == FT_WIRE) {
			glColor3f (0.0f, 0.0f, 0.0f);
			glBegin(GL_LINE_STRIP);
			for (int j=0; j<curface->points.GetSize(); j++) {
				x = (double)((int)vertices[3*curface->points[j]]);
				y = (double)((int)vertices[3*curface->points[j]+1]);
				z = -(double)((int)vertices[3*curface->points[j]+2]);
				x /= 1000;
				y /= 1000;
				z /= 1000;
				tx = (double)(curface->tex_x[j])/(double)(0xFF0000);
				ty = (double)(curface->tex_y[j])/(double)(0xFF0000);

				x *= zoom; y *= zoom; z *= zoom;
				
				glVertex3d(x, y, z);
			}
			j = 0;
			x = (double)((int)vertices[3*curface->points[j]]);
			y = (double)((int)vertices[3*curface->points[j]+1]);
			z = -(double)((int)vertices[3*curface->points[j]+2]);
			x /= 1000;
			y /= 1000;
			z /= 1000;
			tx = (double)(curface->tex_x[j])/(double)(0xFF0000);
			ty = (double)(curface->tex_y[j])/(double)(0xFF0000);

			x *= zoom; y *= zoom; z *= zoom;
			
			glVertex3d(x, y, z);
			glEnd();
		}	// if FT_WIRE
	}

	///////////////////////////////////////////
	//  DRAW CURRENT FACE
	if (current_face > -1) {

		glColor3f (1.0f, 0.0f, 1.0f); // purple

		curface = (face *)faces[current_face];

		// Texture benutzen...
		csOldtex = curface->texture;
		if (csOldtex.Left(5)=="ANIM:") {
			TRACE("Animated Texture: %s\n",csOldtex);
			TRACE("TexSize : %d\n",curface->tex_size);
			csOldtex=GetFirstAnimTexture(csOldtex.Right(csOldtex.GetLength()-5));
		}

		glTexImage2D(GL_TEXTURE_2D, 0, 3, curface->tex_size, 
					curface->tex_size, 0, GL_RGB, GL_UNSIGNED_BYTE, 
  					GetTexturePointer(csOldtex));

		if (curface->points.GetSize()==3)
			glBegin(GL_TRIANGLES);
		else if (curface->points.GetSize()==4)
			glBegin(GL_QUADS);

		for (int j=0; j<curface->points.GetSize(); j++) {
			x = (double)((int)vertices[3*curface->points[j]]);
			y = (double)((int)vertices[3*curface->points[j]+1]);
			z = -(double)((int)vertices[3*curface->points[j]+2]);
			x /= 1000;
			y /= 1000;
			z /= 1000;
			tx = (double)(curface->tex_x[j])/(double)(0xFF0000);
			ty = (double)(curface->tex_y[j])/(double)(0xFF0000);

			x *= zoom; y *= zoom; z *= zoom;

			// Looks better w/o normal vector
			// glNormal3i(curface->normal[0],curface->normal[1],curface->normal[2]);
			glTexCoord2f(tx, ty);
			glVertex3d(x, y, z);
		}
		glEnd();
	}

	///////////////////////////////////////////
	//  DRAW VERTICES

	if (bVerts) {
		glBegin(GL_POINTS);
		for (i=0; i<GetNumberOfVertices(); i++) {
			glColor3f (0.0f, 0.0f, 0.0f);
			if (vertices_tag[i]) glColor3f (1.0f, 1.0f, 1.0f);
			
			x = (double)((int)vertices[3*i]);
			y = (double)((int)vertices[3*i+1]);
			z = -(double)((int)vertices[3*i+2]);
			x /= 1000;
			y /= 1000;
			z /= 1000;

			glVertex3d(zoom*x, zoom*y, zoom*z);
		}
		glEnd();

		///////////////////////////////////////////
		//  ENHANCE TAGGED VERTICES
		for (i=0; i<GetNumberOfVertices(); i++) {
			if (!vertices_tag[i]) continue;
			
			x = (double)((int)vertices[3*i]);
			y = (double)((int)vertices[3*i+1]);
			z = -(double)((int)vertices[3*i+2]);
			x /= 1000;
			y /= 1000;
			z /= 1000;

			glTranslated(zoom*x, zoom*y, zoom*z);
			auxSolidSphere(0.5);
			glTranslated(-zoom*x, -zoom*y, -zoom*z);

		}
	}

	///////////////////////////////////////////
	//  DRAW COORDCROSS FOR CURRENT VERTEX
	if (current_vertex > -1) {
		x = (double)((int)vertices[3*current_vertex]);
		y = (double)((int)vertices[3*current_vertex+1]);
		z = -(double)((int)vertices[3*current_vertex+2]);
		x /= 1000;
		y /= 1000;
		z /= 1000;

		glBegin(GL_LINES);
		glVertex3d(zoom*x, zoom*y, -1000.0);
		glVertex3d(zoom*x, zoom*y, +1000.0);
		glVertex3d(zoom*x, -1000.0, zoom*z);
		glVertex3d(zoom*x, +1000.0, zoom*z);
		glVertex3d(-1000.0, zoom*y, zoom*z);
		glVertex3d(+1000.0, zoom*y, zoom*z);
		glEnd();
	}

	if (bAxes) {
		// Draw Coord Lines
		glBegin(GL_LINES);
		glColor3f (0.0f, 0.0f, 1.0f);
		glVertex3d(0, 0, -1000.0);
		glVertex3d(0, 0, +1000.0);
		glColor3f (0.0f, 1.0f, 0.0f);
		glVertex3d(0, -1000.0, 0);
		glVertex3d(0, +1000.0, 0);
		glColor3f (1.0f, 0.0f, 0.0f);
		glVertex3d(-1000.0, 0, 0);
		glVertex3d(+1000.0, 0, 0);
		glEnd();

		// Mark Directions
		static double sdXMax, sdYMax, sdZMax;
		sdXMax = GetMaxCoord("x") * 1.2;
		sdYMax = GetMaxCoord("y") * 1.2;
		sdZMax = GetMaxCoord("z") * 1.2;

		sdXMax *= zoom/1000.0;
		sdYMax *= zoom/1000.0;
		sdZMax *= zoom/1000.0;

		glColor3f (1.0f, 0.0f, 0.0f);
		glTranslated(sdXMax, 0.0, 0.0);
		auxSolidSphere(0.5);

		glColor3f (0.0f, 1.0f, 0.0f);
		glTranslated(-sdXMax, sdYMax, 0.0);
		auxSolidSphere(0.5);

		glColor3f (0.0f, 0.0f, 1.0f);
		glTranslated(0.0, -sdYMax, -sdZMax);	// z gespiegelt!
		auxSolidSphere(0.5);
		glTranslated(0.0, 0.0, sdZMax);
	}

	/* glEndList();
	bRebuildDisplayList = false; */
}


int CBINModel::LoadTexture(CString name, CString csARTDir)
{
	int i, j;
	int tex_size;
	CFile texfile;

	name.MakeUpper();

	// check if already loaded
	if (cmapTexSizes.Lookup(name, tex_size)) return tex_size;

	if (!texfile.Open(csARTDir+"\\"+name, CFile::modeRead, NULL)) {
		AfxMessageBox("Error loading BIN:\nCould not find the texture '"+name+"'\nin the folder '"+csARTDir+"'!");
		return -1;
	}

	int filelen = texfile.GetLength();
	if (filelen==64*64) tex_size = 64;
	else if (filelen==256*256) tex_size = 256;

	TRACE("CBINModel::LoadTexture(%s, %s)\n", name, csARTDir);

	BYTE *tex, pix, *buf, *bmp;

	tex = (BYTE *)malloc(3*tex_size * tex_size);
	bmp = (BYTE *)malloc(3*tex_size * tex_size);
	buf = (BYTE *)malloc(tex_size * tex_size);

	texfile.Read(buf, filelen);
	texfile.Close();

	// look for custom palette
	CString csCusPal = csARTDir +"\\"+ CopTools::Before(name, ".RAW") + ".ACT";
	TRACE("CBINModel::LoadTexture() : Looking for custom ACT '%s'\n",csCusPal);
	
	if (CopTools::FileExists(csCusPal)) {
		unsigned char cuspal[3*256];

		CFile pal;

		pal.Open(csCusPal, CFile::modeRead, NULL); 
		pal.Read(cuspal, 3*256);
		pal.Close();

		for (i=0; i<filelen; i++) {
			tex[3*i] = cuspal[3*buf[i]];
			tex[3*i+1] = cuspal[3*buf[i]+1];
			tex[3*i+2] = cuspal[3*buf[i]+2];
		}
	}
	else {
		for (i=0; i<filelen; i++) {
			tex[3*i] = palette[3*buf[i]];
			tex[3*i+1] = palette[3*buf[i]+1];
			tex[3*i+2] = palette[3*buf[i]+2];
		}
	}

	int joff0, joff1, joff2, tsi, ti;
	for (j=0; j<tex_size; j++) {
		joff0 = 3*tex_size*j;
		joff1 = 3*tex_size*j+1;
		joff2 = 3*tex_size*j+2;
		for (i=0; i<tex_size; i++) {
			tsi = 3*(tex_size-1-i);
			ti  = 3*i;
			bmp[joff0 + tsi] = tex[joff2 + ti];
			bmp[joff1 + tsi] = tex[joff1 + ti];
			bmp[joff2 + tsi] = tex[joff0 + ti];
		}
	}

	csaBINTextures.Add(name);
	cmapTexPtrs[name] = tex;
	cmapBmpTexPtrs[name] = bmp;
	cmapTexSizes[name] = tex_size;

	free(buf);

	TRACE("CBINModel::LoadTexture() = %d\n", tex_size);
	return tex_size;
}		

BYTE * CBINModel::GetTexturePointer(CString name)
{
	name.MakeUpper();
	//TRACE("CBINModel::GetTexturePointer(%s)\n", name);
	
	void *pBuf;
	if (cmapTexPtrs.Lookup(name, pBuf)) return (BYTE *)pBuf;
	
	//TRACE("CBINModel::GetTexturePointer() - Unknown texture\n", name);
	return NULL;
}

BYTE * CBINModel::GetBmpTexturePointer(CString name)
{
	name.MakeUpper();

	void *pBuf;
	if (cmapBmpTexPtrs.Lookup(name, pBuf)) return (BYTE *)pBuf;
	
	return NULL;
}

int CBINModel::GetTextureSize(CString name)
{
	TRACE("CBINModel::GetTextureSize(%s)\n", name);
	int iSize;
	if (cmapTexSizes.Lookup(name, iSize)) return iSize;

	if(name.Left(5)=="ANIM:") {
		CString csFirst = GetFirstAnimTexture(name.Right(name.GetLength()-5));
		TRACE("CBINModel::GetTextureSize() - ANIM:%s\n", csFirst);
		if (cmapTexSizes.Lookup(csFirst, iSize)) return iSize;
	}

	TRACE("CBINModel::GetTextureSize() - No Success!\n");
	return -1;
}

void CBINModel::ASCIIDump(CString filename)
{
	char buf[100];
	CStdioFile dump;
	long int nx, ny, nz;

	dump.Open(filename, CFile::modeCreate + CFile::modeWrite, NULL);

	dump.WriteString("Model Name\n");
	dump.WriteString(CString(name)+"\n");
	dump.WriteString("Number of Vertices\n");
	dump.WriteString(CString(itoa(GetNumberOfVertices(), buf, 10))+"\n");
	for (int i=0; i<GetNumberOfVertices(); i++) {
		dump.WriteString(	CString(itoa(vertices[3*i+0], buf, 10))+","+
							CString(itoa(vertices[3*i+1], buf, 10))+","+
							CString(itoa(vertices[3*i+2], buf, 10))+"\n");
	}

	face *curface;
	for (i=0; i<GetNumberOfFaces(); i++) {
		curface = (face *)faces[i];

		dump.WriteString("Texture Name\n");
		dump.WriteString(curface->texture+"\n");
		dump.WriteString("Face Type\n");
		dump.WriteString(CString(itoa(curface->type, buf, 10))+"\n");
		dump.WriteString("Normal Vector\n");
		dump.WriteString(	CString(itoa(curface->normal[0], buf, 10))+","+
							CString(itoa(curface->normal[1], buf, 10))+","+
							CString(itoa(curface->normal[2], buf, 10))+"\n");
		dump.WriteString("Calculated Normal Vector\n");
		GetFaceNormal(i, nx, ny, nz, TRUE);	// get normalized normal 8)
		dump.WriteString(	CString(itoa(nx, buf, 10))+","+
							CString(itoa(ny, buf, 10))+","+
							CString(itoa(nz, buf, 10))+"\n");
		dump.WriteString("Magic Number\n");
		dump.WriteString(CString(itoa(curface->unknown, buf, 10))+"\n");
		dump.WriteString(CString(itoa(curface->unknown, buf, 16))+"\n");
		dump.WriteString("Number of Vertices\n");
		dump.WriteString(CString(itoa(curface->points.GetSize(), buf, 10))+"\n");
		for (int j=0; j<curface->points.GetSize(); j++) {
			dump.WriteString(	CString(itoa(curface->points[j], buf, 10))+" ("+
								CString(itoa(curface->tex_x[j], buf, 10))+","+
								CString(itoa(curface->tex_y[j], buf, 10))+")\n");
		}
	}

	dump.Close();
}


void CBINModel::WriteBIN(CString filename, BOOL bBackdrop, BOOL bVertNorm, CWordArray &cwaTrans, long int liHellMagic)
// CString filename		- name of the destination file
// BOOL bBackdrop		- TRUE if model will be used as backdrop 
//							do not recalculate magic number
// BOOL bVertNorm       - write 0x03 Vertex Normal Block
// CWordArray &cwaTrans - log changes in face numbering to convert face selections on saving
// long int liHellMagic - yet unknown Hellbender number in header
{
	int i;
	int iFaceNum = -1;
	
	cwaTrans.RemoveAll();
	cwaTrans.SetSize(GetNumberOfFaces());

	// Open the BIN file
	CFile bin;

	bin.Open(filename, CFile::modeWrite + CFile::modeCreate, NULL);

	int aux;
	face *curface;

	// write BIN header
	aux = 0x00000014;  bin.Write(&aux, 4);
	aux = liHellMagic; bin.Write(&aux, 4);
	aux = 0x00000002;  bin.Write(&aux, 4);
	aux = 0x00000000;  bin.Write(&aux, 4);

	// write number of vertices
	aux = GetNumberOfVertices();
	bin.Write(&aux, 4);

	// write vertex coords
	for (i=0; i<GetNumberOfVertices(); i++) {
		aux = vertices[3*i+0]; bin.Write(&aux, 4);
		aux = vertices[3*i+1]; bin.Write(&aux, 4);
		aux = vertices[3*i+2]; bin.Write(&aux, 4);
	}

	// if desired, write vertex normals
	if (bVertNorm) {
		aux = 0x00000003; bin.Write(&aux, 4);
		aux = 0x00000000; bin.Write(&aux, 4);

		// write number of vertices
		aux = GetNumberOfVertices();
		bin.Write(&aux, 4);

		long int lx, ly, lz;

		for (i=0; i<GetNumberOfVertices(); i++) {
			GetVertexNormal(i, lx, ly, lz);
			bin.Write(&lx, 4);
			bin.Write(&ly, 4);
			bin.Write(&lz, 4);
		}
	}

	// now write the "Texture Blocks"
	for (i=0; i<csaBINTextures.GetSize(); i++) {
		if (!IsTextureInUse(csaBINTextures[i])) continue;

		aux = 0x0000000D; bin.Write(&aux, 4);
		aux = 0x00000000; bin.Write(&aux, 4);
		bin.Write((LPCTSTR)csaBINTextures[i], 16);

		// write all faces using this particular texture...
		for (int j=0; j<GetNumberOfFaces(); j++) {
			long int nx, ny, nz;
			GetFaceNormal(j, nx, ny, nz, TRUE);

			curface = (face *)faces[j];
			if (curface->texture == csaBINTextures[i]) {

				iFaceNum++;				// new face index
				cwaTrans[j] = iFaceNum;	// log changes (for changing the Face Selections)

				aux = curface->type; bin.Write(&aux, 4);
				aux = curface->points.GetSize(); bin.Write(&aux, 4);
				bin.Write(&nx, 4);
				bin.Write(&ny, 4);
				bin.Write(&nz, 4);
				
				if (!bBackdrop) aux = CalcFaceMagic(j);
				else aux = curface->unknown;
				bin.Write(&aux, 4);

				for (int k=0; k<curface->points.GetSize(); k++) {
					aux = curface->points[k]; bin.Write(&aux, 4);
					aux = curface->tex_x[k]; bin.Write(&aux, 4);
					aux = curface->tex_y[k]; bin.Write(&aux, 4);
				}	// for "curface points"
			}	// curface->texture == tex_names[i]
		}	// for "vertices"
	}	// for "textures"

	// now write the "Texture Animation Blocks"
	sTexAnim *psTA;
	CString csAnim;
	for (i=0; i<cpaTexAnims.GetSize(); i++) {
		psTA = (sTexAnim *)cpaTexAnims[i];
		csAnim = psTA->csName;

		if (!IsTextureInUse("ANIM:"+csAnim)) continue;

		// write texture header
		aux = 0x0000001D; bin.Write(&aux, 4);
		aux = 0x00000000; bin.Write(&aux, 4);
		aux = GetAnimNumFrames(csAnim);  bin.Write(&aux, 4);
		aux = 0x00000000; bin.Write(&aux, 4);
		DWORD dwAux = GetAnimDelay(csAnim);  bin.Write(&dwAux, 4);
		TRACE("Delay: %ld\n", GetAnimDelay(csAnim));
		aux = 0x00000000; bin.Write(&aux, 4);
		aux = 0x00000000; bin.Write(&aux, 4);
		for (int iA=0; iA<GetAnimNumFrames(csAnim); iA++)
			bin.Write((LPCTSTR)GetAnimFrame(csAnim, iA), 32);

		// write all faces using this particular texture...
		for (int j=0; j<GetNumberOfFaces(); j++) {

			iFaceNum++;				// new face index
			cwaTrans[j] = iFaceNum;	// log changes (for changing the Face Selections)
			
			long int nx, ny, nz;
			GetFaceNormal(j, nx, ny, nz, TRUE);

			curface = (face *)faces[j];
			if (curface->texture == "ANIM:"+csAnim) {
				aux = curface->type; bin.Write(&aux, 4);
				aux = curface->points.GetSize(); bin.Write(&aux, 4);
				bin.Write(&nx, 4);
				bin.Write(&ny, 4);
				bin.Write(&nz, 4);
				
				if (!bBackdrop) aux = CalcFaceMagic(j);
				else aux = curface->unknown;
				bin.Write(&aux, 4);

				for (int k=0; k<curface->points.GetSize(); k++) {
					aux = curface->points[k]; bin.Write(&aux, 4);
					aux = curface->tex_x[k]; bin.Write(&aux, 4);
					aux = curface->tex_y[k]; bin.Write(&aux, 4);
				}	// for "curface points"
			}	// curface->texture == tex_names[i]
		}	// for "vertices"
	}	// for "textures"

	// close with a zero!
	aux = 0x00000000; bin.Write(&aux, 4);

	bin.Close();
}

BOOL CBINModel::SetCurrentFace(int number)
{
	if (GetNumberOfFaces() == 0) return FALSE;

	if (number < -1) number += GetNumberOfFaces() +1;
	current_face = number % GetNumberOfFaces();

	return TRUE;
}

BOOL CBINModel::SetCurrentVertex(int number)
{
	if (GetNumberOfVertices() == 0) return FALSE;

	if (number < -1) number += GetNumberOfVertices() +1;
	current_vertex = number % (GetNumberOfVertices());

	return TRUE;
}

void CBINModel::SetZoom(double z)
{
	zoom = z;
}

double CBINModel::GetZoom(void)
{
	return zoom;
}

int CBINModel::GetCurrentVertex(void)
{
	return current_vertex;
}

int CBINModel::GetCurrentFace(void)
{
	return current_face;
}

int CBINModel::GetHellMagic(void)
{
	return liHellMagic;
}

int CBINModel::CrossProduct(double ax, double ay, double az, double bx, double by, double bz, double &cx, double &cy, double &cz) 
{
	cx = ay*bz - az*by;
	cy = az*bx - ax*bz;
	cz = ax*by - ay*bx;
	return 1;	
}

void CBINModel::DrawSideViewX(CDC *pDC, int x0, int y0, int w, int h)
{
	int i;
	double y, z;
	face *curface;

	CPen *oldpen;
	CPen black(PS_SOLID, 1, RGB(0,0,0));
	CPen red(PS_SOLID, 1, RGB(255,0,0));
	CPen green(PS_SOLID, 1, RGB(0,255,0));
	CPen blue(PS_SOLID, 1, RGB(0,0,255));
	CPen white(PS_SOLID, 1, RGB(255,255,255));

	oldpen = pDC->SelectObject(&black);

	int miny = GetMinCoord("y");
	int maxy = GetMaxCoord("y");
	int minz = GetMinCoord("z");
	int maxz = GetMaxCoord("z");

	double fac = min(w/(double)(maxz-minz), h/(double)(maxy-miny));

	// draw all faces
	for (i=0; i<GetNumberOfFaces(); i++) {

		if (i == current_face) 	continue;

		pDC->SelectObject(&black);
		if (faces_tag[i]) pDC->SelectObject(&green);

		curface = (face *)faces[i];

		for (int j=0; j<curface->points.GetSize(); j++) {
			y = (double)((int)vertices[3*curface->points[j]+1]);
			z = (double)((int)vertices[3*curface->points[j]+2]);

			y -= miny;
			z -= minz;

			y *= fac;
			z *= fac;

			if (j==0) {
				pDC->MoveTo(x0+z,y0+h-y);
			} else {
				pDC->LineTo(x0+z,y0+h-y);
			}
		}
		j = 0;
		y = (double)((int)vertices[3*curface->points[j]+1]);
		z = (double)((int)vertices[3*curface->points[j]+2]);

		y -= miny;
		z -= minz;

		y *= fac;
		z *= fac;
		
		pDC->LineTo(x0+z,y0+h-y);
	}

	if (current_face > -1) {
		pDC->SelectObject(&white);
		i = current_face;
		curface = (face *)faces[i];

		for (int j=0; j<curface->points.GetSize(); j++) {
			y = (double)((int)vertices[3*curface->points[j]+1]);
			z = (double)((int)vertices[3*curface->points[j]+2]);

			y -= miny;
			z -= minz;

			y *= fac;
			z *= fac;

			if (j==0) {
				pDC->MoveTo(x0+z,y0+h-y);
			} else {
				pDC->LineTo(x0+z,y0+h-y);
			}
		}
		j = 0;
		y = (double)((int)vertices[3*curface->points[j]+1]);
		z = (double)((int)vertices[3*curface->points[j]+2]);

		y -= miny;
		z -= minz;

		y *= fac;
		z *= fac;
		
		pDC->LineTo(x0+z,y0+h-y);
	}
	
	if (current_vertex > -1) {
		pDC->SelectObject(&red);

		y = (double)((int)vertices[3*current_vertex+1]);
		z = (double)((int)vertices[3*current_vertex+2]);

		y -= miny;
		z -= minz;

		y *= fac;
		z *= fac;

		pDC->Ellipse(x0+z-2,y0+h-y-2,x0+z+2,y0+h-y+2);
	}

	pDC->SelectObject(oldpen);
}

void CBINModel::DrawSideViewY(CDC *pDC, int x0, int y0, int w, int h)
{
	int i;
	double x, z;
	face *curface;

	CPen *oldpen;
	CPen black(PS_SOLID, 1, RGB(0,0,0));
	CPen red(PS_SOLID, 1, RGB(255,0,0));
	CPen blue(PS_SOLID, 1, RGB(0,0,255));
	CPen green(PS_SOLID, 1, RGB(0,255,0));
	CPen white(PS_SOLID, 1, RGB(255,255,255));

	oldpen = pDC->SelectObject(&black);

	int minx = GetMinCoord("x");
	int maxx = GetMaxCoord("x");
	int minz = GetMinCoord("z");
	int maxz = GetMaxCoord("z");

	double fac = min(w/(double)(maxx-minx), h/(double)(maxz-minz));

	// draw all faces
	for (i=0; i<GetNumberOfFaces(); i++) {

		if (i == current_face) continue;	// except the "current" face

		pDC->SelectObject(&black);
		if (faces_tag[i]) pDC->SelectObject(&green);

		curface = (face *)faces[i];

		for (int j=0; j<curface->points.GetSize(); j++) {
			x = (double)((int)vertices[3*curface->points[j]+0]);
			z = (double)((int)vertices[3*curface->points[j]+2]);

			x -= minx;
			z -= minz;

			x *= fac;
			z *= fac;

			if (j==0) {
				pDC->MoveTo(x0+x,y0+h-z);
			} else {
				pDC->LineTo(x0+x,y0+h-z);
			}
		}
		j = 0;
		x = (double)((int)vertices[3*curface->points[j]+0]);
		z = (double)((int)vertices[3*curface->points[j]+2]);

		x -= minx;
		z -= minz;

		x *= fac;
		z *= fac;
		
		pDC->LineTo(x0+x,y0+h-z);
	}

	if (current_face > -1) {
		pDC->SelectObject(&white);
		i = current_face;
		curface = (face *)faces[i];
		for (int j=0; j<curface->points.GetSize(); j++) {
			x = (double)((int)vertices[3*curface->points[j]+0]);
			z = (double)((int)vertices[3*curface->points[j]+2]);

			x -= minx;
			z -= minz;

			x *= fac;
			z *= fac;

			if (j==0) {
				pDC->MoveTo(x0+x,y0+h-z);
			} else {
				pDC->LineTo(x0+x,y0+h-z);
			}
		}
		j = 0;
		x = (double)((int)vertices[3*curface->points[j]+0]);
		z = (double)((int)vertices[3*curface->points[j]+2]);

		x -= minx;
		z -= minz;

		x *= fac;
		z *= fac;
		
		pDC->LineTo(x0+x,y0+h-z);
	}

	if (current_vertex > -1) {
		pDC->SelectObject(&red);

		x = (double)((int)vertices[3*current_vertex+0]);
		z = (double)((int)vertices[3*current_vertex+2]);

		x -= minx;
		z -= minz;

		x *= fac;
		z *= fac;
		
		pDC->Ellipse(x0+x-2,y0+h-z-2,x0+x+2,y0+h-z+2);
	}

	pDC->SelectObject(oldpen);
}

void CBINModel::DrawSideViewZ(CDC *pDC, int x0, int y0, int w, int h)
{
	int i;
	double x, y;
	face *curface;

	CPen *oldpen;
	CPen black(PS_SOLID, 1, RGB(0,0,0));
	CPen red(PS_SOLID, 1, RGB(255,0,0));
	CPen blue(PS_SOLID, 1, RGB(0,0,255));
	CPen green(PS_SOLID, 1, RGB(0,255,0));
	CPen white(PS_SOLID, 1, RGB(255,255,255));

	oldpen = pDC->SelectObject(&black);
	
	int minx = GetMinCoord("x");
	int maxx = GetMaxCoord("x");
	int miny = GetMinCoord("y");
	int maxy = GetMaxCoord("y");

	double fac = min(w/(double)(maxx-minx), h/(double)(maxy-miny));

	// draw all faces
	for (i=0; i<GetNumberOfFaces(); i++) {

		if (i == current_face) continue;	// except the "current" face

		pDC->SelectObject(&black);
		if (faces_tag[i]) pDC->SelectObject(&green);

		curface = (face *)faces[i];

		for (int j=0; j<curface->points.GetSize(); j++) {
			x = (double)((int)vertices[3*curface->points[j]+0]);
			y = (double)((int)vertices[3*curface->points[j]+1]);

			x -= minx;
			y -= miny;

			x *= fac;
			y *= fac;

			if (j==0) {
				pDC->MoveTo(x0+x,y0+h-y);
			} else {
				pDC->LineTo(x0+x,y0+h-y);
			}
		}
		j = 0;
		x = (double)((int)vertices[3*curface->points[j]+0]);
		y = (double)((int)vertices[3*curface->points[j]+1]);

		x -= minx;
		y -= miny;

		x *= fac;
		y *= fac;
		
		pDC->LineTo(x0+x,y0+h-y);
	}

	if (current_face > -1) {
		pDC->SelectObject(&white);
		i = current_face;
		curface = (face *)faces[i];

		for (int j=0; j<curface->points.GetSize(); j++) {
			x = (double)((int)vertices[3*curface->points[j]+0]);
			y = (double)((int)vertices[3*curface->points[j]+1]);

			x -= minx;
			y -= miny;

			x *= fac;
			y *= fac;

			if (j==0) {
				pDC->MoveTo(x0+x,y0+h-y);
			} else {
				pDC->LineTo(x0+x,y0+h-y);
			}
		}
		j = 0;
		x = (double)((int)vertices[3*curface->points[j]+0]);
		y = (double)((int)vertices[3*curface->points[j]+1]);

		x -= minx;
		y -= miny;

		x *= fac;
		y *= fac;
		
		pDC->LineTo(x0+x,y0+h-y);
	}

	if (current_vertex > -1) {
		pDC->SelectObject(&red);

		x = (double)((int)vertices[3*current_vertex+0]);
		y = (double)((int)vertices[3*current_vertex+1]);

		x -= minx;
		y -= miny;

		x *= fac;
		y *= fac;
		
		pDC->Ellipse(x0+x-2,y0+h-y-2,x0+x+2,y0+h-y+2);
	}

	pDC->SelectObject(oldpen);
}

int CBINModel::GetMinCoord(CString xyz)
{
	if (GetNumberOfVertices() == 0) return 1;

	int val,i;

	if (xyz == "x") {
		val = (int)vertices[0];
		for (i=0; i<GetNumberOfVertices(); i++) {
			val = min(val, (int)vertices[3*i]);
		}
	} 
	else if (xyz == "y") {
		val = (int)vertices[1];
		for (i=0; i<GetNumberOfVertices(); i++) {
			val = min(val, (int)vertices[3*i+1]);
		}
	} 
	else if (xyz == "z") {
		val = (int)vertices[2];
		for (i=0; i<GetNumberOfVertices(); i++) {
			val = min(val, (int)vertices[3*i+2]);
		}
	}

	return val;
}

int CBINModel::GetMaxCoord(CString xyz)
{
	if (GetNumberOfVertices() == 0) return 1;

	int val,i;

	if (xyz == "x") {
		val = (int)vertices[0];
		for (i=0; i<GetNumberOfVertices(); i++) {
			val = max(val, (int)vertices[3*i]);
		}
	} 
	else if (xyz == "y") {
		val = (int)vertices[1];
		for (i=0; i<GetNumberOfVertices(); i++) {
			val = max(val, (int)vertices[3*i+1]);
		}
	} 
	else if (xyz == "z") {
		val = (int)vertices[2];
		for (i=0; i<GetNumberOfVertices(); i++) {
			val = max(val, (int)vertices[3*i+2]);
		}
	}

	return val;
}

void CBINModel::GetVertex(int num, int &x, int &y, int &z)
{
	if (num<0) return;

	x = (int)vertices[3*num+0];
	y = (int)vertices[3*num+1];
	z = (int)vertices[3*num+2];
}

void CBINModel::SetVertex(int num, int x, int y, int z)
{
	if (num<0) return;

	vertices[3*num+0] = (UINT)x;
	vertices[3*num+1] = (UINT)y;
	vertices[3*num+2] = (UINT)z;
}

void CBINModel::TranslateVertex(int num, int dx, int dy, int dz, BOOL bTaggedOnly)
{
	if (num<0) return;
	if (bTaggedOnly && !IsVertexTagged(num)) return;

	int x = (int)(vertices[3*num+0]) + dx;
	int y = (int)(vertices[3*num+1]) + dy;
	int z = (int)(vertices[3*num+2]) + dz;

	vertices[3*num+0] = (UINT)x;
	vertices[3*num+1] = (UINT)y;
	vertices[3*num+2] = (UINT)z;	
}

void CBINModel::GetFaceNormal(int num, long int &nx, long int &ny, long int &nz, BOOL normalize)
{
	face *curface;
	double x[4], y[4], z[4];
	double vx[4], vy[4], vz[4];

	curface = (face *)faces[num];

	for (int i=0; i<curface->points.GetSize(); i++) {
		x[i] = (long int)vertices[3*curface->points[i]];
		y[i] = (long int)vertices[3*curface->points[i]+1];
		z[i] = (long int)vertices[3*curface->points[i]+2];
	}

	double ax, ay, az;

	if (curface->points.GetSize() == 3) {
		CrossProduct(x[1]-x[0],y[1]-y[0],z[1]-z[0],x[2]-x[0],y[2]-y[0],z[2]-z[0],ax, ay, az);
	}
	else if (curface->points.GetSize() == 4) {
		CrossProduct(x[1]-x[0],y[1]-y[0],z[1]-z[0],x[3]-x[0],y[3]-y[0],z[3]-z[0],vx[0], vy[0], vz[0]);
		CrossProduct(x[2]-x[1],y[2]-y[1],z[2]-z[1],x[0]-x[1],y[0]-y[1],z[0]-z[1],vx[1], vy[1], vz[1]);
		CrossProduct(x[3]-x[2],y[3]-y[2],z[3]-z[2],x[1]-x[2],y[1]-y[2],z[1]-z[2],vx[2], vy[2], vz[2]);
		CrossProduct(x[0]-x[3],y[0]-y[3],z[0]-z[3],x[2]-x[3],y[2]-y[3],z[2]-z[3],vx[3], vy[3], vz[3]);
		
		double lx, ly, lz;
		lx = vx[0] + vx[1] + vx[2] + vx[3];
		ly = vy[0] + vy[1] + vy[2] + vy[3];
		lz = vz[0] + vz[1] + vz[2] + vz[3];
	
		ax = lx/3.0;
		ay = ly/3.0;
		az = lz/3.0;
	}

	// normalize...
	if (normalize) {
		double dx, dy, dz;

		dx = (double)ax;
		dy = (double)ay;
		dz = (double)az;

		double len = sqrt(dx*dx+dy*dy+dz*dz);

		double factor = 65535.0/len;

		nx = (long int)(factor*dx);
		ny = (long int)(factor*dy);
		nz = (long int)(factor*dz);
	}
}

void CBINModel::GetVertexNormal(int num, long int &nx, long int &ny, long int &nz)
{
	int i, cnt;
	long int fx, fy, fz;
	double x, y, z;

	x = y = z = 0.0;
	cnt = 0;

	for (i=0; i<GetNumberOfFaces(); i++) {
		if (IsVertexInFace(num, i)) {
			GetFaceNormal(i, fx, fy, fz, TRUE);	

			x += (double)fx;
			y += (double)fy;
			z += (double)fz;

			cnt++;
		}
	}

	x /= (double)cnt;
	y /= (double)cnt;
	z /= (double)cnt;

	// normalize...
	double len = sqrt(x*x+y*y+z*z);

	double factor = 65535.0/len;

	nx = (long int)(factor*x);
	ny = (long int)(factor*y);
	nz = (long int)(factor*z);
}

void CBINModel::CycleCurrentFaceVertex(BOOL reset)
{
	static int cface, vernum;
	static int oldface = -1;

	cface = GetCurrentFace();
	if (cface == -1) return;

	if (cface != oldface || reset) {
		vernum = 0;
		oldface = cface;
	} else vernum++;


	face *curface = (face *)faces[cface];

	if (vernum >= curface->points.GetSize()) vernum = 0;

	current_vertex = curface->points[vernum];
}

BOOL CBINModel::IsTextureInUse(CString texname)
{ 
	face *curface;
	for (int j=0; j<GetNumberOfFaces(); j++) {
		curface = (face *)faces[j];
		if (curface->texture == texname) {
			return TRUE;
		}
	}
	return FALSE;
}

CString CBINModel::GetCurrentFaceTextureName()
{
	return GetFaceTextureName(GetCurrentFace());
}

void CBINModel::SetFaceTexture(int iFace, CString csTexture)
{
	if (!ValidFace(iFace)) return;
	
	face *curface = (face *)faces[iFace];
	curface->texture = csTexture;
	curface->tex_size = GetTextureSize(csTexture);
}

int CBINModel::GetCurrentFaceTextureSize()
{
	return GetFaceTextureSize(GetCurrentFace());
}

void CBINModel::DrawTextureCoordSquare(int x, int y, int w, int size, int iface, CDC *pDC)
{
	if (iface == -1) return;
	face *curface = (face *)faces[iface];

	double tx, ty;

	CPen *oldpen;
	CPen white(PS_SOLID, 1, RGB(255,255,255));

	oldpen = pDC->SelectObject(&white);

	int iOldROP2 = pDC->SetROP2(R2_XORPEN);

	for (int j=0; j<curface->points.GetSize(); j++) {
		tx = (double)(curface->tex_x[j])/(double)(0xFF0000);
		ty = (double)(curface->tex_y[j])/(double)(0xFF0000);

//		if (size == 64) {
			if (j==0)
				pDC->MoveTo(x+w-w*tx, y+w*ty);
			else
				pDC->LineTo(x+w-w*tx, y+w*ty);
			if (curface->points[j] == current_vertex) {
				pDC->Ellipse(x+w-w*tx-3, y+w*ty-3, x+w-w*tx+3, y+w*ty+3);
			}
//		}
//		else if (size == 256) {
//			if (j==0)
//				pDC->MoveTo(x+w*tx, y+w*ty);
//			else
//				pDC->LineTo(x+w*tx, y+w*ty);
//			if (curface->points[j] == current_vertex) {
//				pDC->Ellipse(x+w*tx-3, y+w*ty-3, x+w*tx+3, y+w*ty+3);
//			}
//		}
	}
	tx = (double)(curface->tex_x[0])/(double)(0xFF0000);
	ty = (double)(curface->tex_y[0])/(double)(0xFF0000);
//		if (size == 64) 
			pDC->LineTo(x+w-w*tx, y+w*ty);
//		else if (size == 256) 
//			pDC->LineTo(x+w*tx, y+w*ty);

	pDC->SetROP2(iOldROP2);
	pDC->SelectObject(oldpen);
}

CString CBINModel::SelectTexture()
{
	CTexListDlg dlg(NULL, &csaBINTextures, &cpaTexAnims);
	dlg.bin = this;

	if (dlg.DoModal() == IDOK) {
		return dlg.m_selection;
	}
	else return "";
}

void CBINModel::SetCurrentFaceTexture(CString name)
{
	if (current_face == -1) return;
	if (name.IsEmpty()) return;
	SetFaceTexture(GetCurrentFace(), name);
}

void CBINModel::MoveCurrentVertexTextureCoords(int dx, int dy)
{
	if (current_face == -1) return;

	face *curface = (face *)faces[current_face];

	int num;
	for (num=0; curface->points[num] != current_vertex; num++);

	double tx = (double)(curface->tex_x[num])/(double)(0xFF0000);
	double ty = (double)(curface->tex_y[num])/(double)(0xFF0000);
	tx = (int)(255.0 * tx);
	ty = (int)(255.0 * ty);

	tx += dx;
	ty +=dy;

	if (tx < 0) tx = 0;
	if (ty < 0) ty = 0;
	if (tx > 255) tx = 255;
	if (ty > 255) ty = 255;

	tx = (tx * (double)(0xFF0000))/255.0;
	ty = (ty * (double)(0xFF0000))/255.0;

	curface->tex_x[num] = (int)tx;
	curface->tex_y[num] = (int)ty;
}

void CBINModel::SetCurrentVertexTextureCoords(double tx, double ty)
{
	if (current_face == -1) return;

// if (GetCurrentFaceTextureSize() == 64) {
//		tx = 1.0 - tx;
//	}

	face *curface = (face *)faces[current_face];

	int num;
	for (num=0; curface->points[num] != current_vertex; num++);

	tx = ((double)tx * (double)(0xFF0000));
	ty = ((double)ty * (double)(0xFF0000));

	curface->tex_x[num] = (int)tx;
	curface->tex_y[num] = (int)ty;
}

CString CBINModel::GetInfoString()
{
	char buf[200];
	CString info = "";

	////////////////
	// GENERAL INFO
	info += "MODEL INFORMATION:\n";
	info += "  Number of vertices.\t"+CString(itoa(GetNumberOfVertices(), buf, 10))+"\n";
	info += "  Number of faces....\t"+CString(itoa(GetNumberOfFaces(), buf, 10))+"\n";
	info += "  Number of textures.\t"+CString(itoa(GetNumberOfUsedTextures(), buf, 10))+"\n";
	info += "  Size x: "+CString(itoa(GetMaxCoord("x")-GetMinCoord("x"), buf, 10))+" units\n";
	info += "  Size y: "+CString(itoa(GetMaxCoord("y")-GetMinCoord("y"), buf, 10))+" units\n";
	info += "  Size z: "+CString(itoa(GetMaxCoord("z")-GetMinCoord("z"), buf, 10))+" units\n";
	
	if (current_face != -1) {
		face *curface = (face *)faces[current_face];
		info += "\n";
		info += "FACE INFORMATION FOR FACE "+CString(itoa(current_face, buf, 10))+"\n";
		info += "  Vertices: (";
		for (int i=0; i<curface->points.GetSize(); i++) {
			if (i<curface->points.GetUpperBound())
				info += CString(itoa(curface->points[i], buf, 10))+", ";
			else
				info += CString(itoa(curface->points[i], buf, 10))+")\n";
		}
		long int xn, yn, zn;
		GetFaceNormal(current_face, xn, yn, zn, TRUE);
		info += "  Normal vector (calculated by BINedit): (";
			info += CString(itoa(xn, buf, 10))+", ";
			info += CString(itoa(yn, buf, 10))+", ";
			info += CString(itoa(zn, buf, 10))+")\n";
		info += "  Normal vector (as loaded):";
			info += CString(itoa(curface->normal[0], buf, 10))+", ";
			info += CString(itoa(curface->normal[1], buf, 10))+", ";
			info += CString(itoa(curface->normal[2], buf, 10))+")\n";
		info += "  Magic value (calculated by BINedit): (";
			info += CString(itoa(CalcFaceMagic(GetCurrentFace()), buf, 10))+")\n";
		info += "  Magic value (as loaded):";
			info += CString(itoa(curface->unknown, buf, 10))+"\n";
		info += "  Texture: "+curface->texture+"("+CString(itoa(curface->tex_size, buf, 10))+")\n";
	}

	if (current_vertex != -1) {
		info += "\n";
		info += "VERTEX INFORMATION FOR VERTEX "+CString(itoa(current_vertex, buf, 10))+"\n";
		info += "  Position: (";
			info += CString(itoa(vertices[3*current_vertex+0], buf, 10))+", ";
			info += CString(itoa(vertices[3*current_vertex+1], buf, 10))+", ";
			info += CString(itoa(vertices[3*current_vertex+2], buf, 10))+")\n";
	}

	return info;
}

int CBINModel::CalcFaceMagic(int fnum)
{
	if (!ValidFace(fnum)) return 0;
	
	long int nx, ny, nz;
	int vx, vy, vz;
	
	face *curface = (face *)faces[fnum];
	vx = vertices[3*curface->points[0]+0];
	vy = vertices[3*curface->points[0]+1];
	vz = vertices[3*curface->points[0]+2];

	GetFaceNormal(fnum, nx, ny, nz, TRUE);
	return  nx*vx + ny*vy + nz*vz;
}

void CBINModel::RotateX(double deg)
{
	double y, z, ry, rz;

	deg = 3.14159265*(deg/180.0);
	for (int i=0; i<GetNumberOfVertices(); i++) {
		y = (int)vertices[3*i+1];
		z = (int)vertices[3*i+2];

		ry = y*cos(deg) - z*sin(deg);
		rz = y*sin(deg) + z*cos(deg);

		vertices[3*i+1] = ry;
		vertices[3*i+2] = rz;
	}
}

void CBINModel::RotateY(double deg)
{
	double x, z, rx, rz;

	deg = 3.14159265*(deg/180.0);
	for (int i=0; i<GetNumberOfVertices(); i++) {
		x = (int)vertices[3*i+0];
		z = (int)vertices[3*i+2];

		rx = x*cos(deg) - z*sin(deg);
		rz = x*sin(deg) + z*cos(deg);

		vertices[3*i+0] = rx;
		vertices[3*i+2] = rz;
	}
}

void CBINModel::RotateZ(double deg)
{
	double x, y, rx, ry;

	deg = 3.14159265*(deg/180.0);
	for (int i=0; i<GetNumberOfVertices(); i++) {
		x = (int)vertices[3*i+0];
		y = (int)vertices[3*i+1];

		rx = x*cos(deg) - y*sin(deg);
		ry = x*sin(deg) + y*cos(deg);

		vertices[3*i+0] = rx;
		vertices[3*i+1] = ry;
	}
}


void CBINModel::Translate(int dx, int dy, int dz, BOOL bTaggedOnly)
{
	for (int i=0; i<GetNumberOfVertices(); i++)
		TranslateVertex(i, dx, dy, dz, bTaggedOnly);
}


int CBINModel::AddFace(int v1, int v2, int v3, int v4, int type)
{
	int i;
	int verts = 4;
	int vertmax = GetNumberOfVertices();

	if (v1<0  || v1>vertmax) return -1;
	if (v2<0  || v2>vertmax) return -1;
	if (v3<0  || v3>vertmax) return -1;
	if (v4<-1 || v4>vertmax) return -1;
	if (v4 == -1) verts = 3;	// triangle

	face *nf = new face;

	nf->type = type;
	nf->tex_size = GetTextureSize(csaBINTextures[0]);
	nf->texture = csaBINTextures[0];
	nf->points.Add(v1);
	nf->tex_x.Add(0);
	nf->tex_y.Add(0);
	nf->points.Add(v2);
	nf->tex_x.Add(0xFF0000);
	nf->tex_y.Add(0);
	nf->points.Add(v3);
	nf->tex_x.Add(0xFF0000);
	nf->tex_y.Add(0xFF0000);
	if (verts == 4) {
		nf->points.Add(v4);
		nf->tex_x.Add(0);
		nf->tex_y.Add(0xFF0000);
	}
	faces.Add(nf);
	faces_tag.Add(0);	// new face not tagged

	GetFaceNormal(faces.GetUpperBound(), nf->normal[0], nf->normal[1], nf->normal[2], TRUE);
	nf->unknown = CalcFaceMagic(faces.GetUpperBound());

	return faces.GetUpperBound();
}

int CBINModel::DeleteFace(int fnum)
{
	if (fnum<0 || fnum>faces.GetUpperBound()) return -1;

	faces.RemoveAt(fnum, 1);
	if (current_face > faces.GetUpperBound()) current_face = -1;

	FaceGroupDeleteFace(fnum);
	return 1;
}

int CBINModel::GetNumberOfUsedTextures()
{
	int retval = 0;
	for (int i=0; i<csaBINTextures.GetSize(); i++) {
		if (IsTextureInUse(csaBINTextures[i])) retval++;
	}

	return retval;
}

int CBINModel::GetVertexCoords(int vnum, int &x, int &y, int &z)
{
	if (vnum<0 || vnum>GetNumberOfVertices()) return -1;

	x = vertices[3*vnum+0];
	y = vertices[3*vnum+1];
	z = vertices[3*vnum+2];

	return 1;
}

int CBINModel::SetVertexCoords(int vnum, int x, int y, int z)
{
	if (vnum<0 || vnum>GetNumberOfVertices()) return -1;

	vertices[3*vnum+0] = x;
	vertices[3*vnum+1] = y;
	vertices[3*vnum+2] = z;

	return 1;
}

int CBINModel::AddVertex(int nx, int ny, int nz)
{
	vertices.Add((UINT)nx);
	vertices.Add((UINT)ny);
	vertices.Add((UINT)nz);

	vertices_tag.Add(0);	// new vertex not tagged

	return 1;
}

int CBINModel::Scale(double fx, double fy, double fz)
{
	double x, y, z;

	for (int i=0; i<GetNumberOfVertices(); i++) {
		x = (int)vertices[3*i+0];
		y = (int)vertices[3*i+1];
		z = (int)vertices[3*i+2];

		x = x*fx;
		y = y*fy;
		z = z*fz;

		vertices[3*i+0] = x;
		vertices[3*i+1] = y;
		vertices[3*i+2] = z;
	}
	return 1;
}

int CBINModel::GetNumberOfTextures()
{
	return csaBINTextures.GetSize();
}

void CBINModel::ToggleFaceTag(int num)
{
	if (num<0 || num>GetNumberOfFaces()) return;
	faces_tag[num] = !faces_tag[num];
}

void CBINModel::ToggleVertexTag(int num)
{
	if (num<0 || num>GetNumberOfVertices()) return;
	vertices_tag[num] = !vertices_tag[num];
}

void CBINModel::SaveFaceGroup(CString name)
{
	int i, j;
	CUIntArray *par;

	for (i=0; i<face_tag_names.GetSize(); i++) {
		if (face_tag_names[i] == name) {
			// overwrite existing
			par = (CUIntArray *)face_tag_arrays[i];
			par->RemoveAll();
			for (j=0; j<GetNumberOfFaces(); j++)
				if (faces_tag[j]) par->Add(j);
			return;
		}
	}

	par = new CUIntArray;
	for (j=0; j<GetNumberOfFaces(); j++)
		if (faces_tag[j]) par->Add(j);
	face_tag_names.Add(name);
	face_tag_arrays.Add(par);
}

void CBINModel::SaveVertexGroup(CString name)
{
	int i, j;
	CUIntArray *par;

	for (i=0; i<vertex_tag_names.GetSize(); i++) {
		if (vertex_tag_names[i] == name) {
			// overwrite existing
			par = (CUIntArray *)vertex_tag_arrays[i];
			par->RemoveAll();
			for (j=0; j<GetNumberOfVertices(); j++)
				if (vertices_tag[j]) par->Add(j);
			return;
		}
	}

	par = new CUIntArray;
	for (j=0; j<GetNumberOfVertices(); j++)
		if (vertices_tag[j]) par->Add(j);
	vertex_tag_names.Add(name);
	vertex_tag_arrays.Add(par);
}

void CBINModel::RestoreFaceGroup(CString name, BOOL bAdd)
{
	int i, j;
	CUIntArray *par;

	if (!bAdd) {
		for (j=0; j<GetNumberOfFaces(); j++)
			faces_tag[j] = 0;
	}

	for (i=0; i<face_tag_names.GetSize(); i++) {
		if (face_tag_names[i] == name) {
			// load
			par = (CUIntArray *)face_tag_arrays[i];
			for (j=0; j<par->GetSize(); j++)
				faces_tag[par->GetAt(j)] = 1;
			return;
		}
	}
}

void CBINModel::RestoreVertexGroup(CString name, BOOL bAdd)
{
	int i, j;
	CUIntArray *par;

	if (!bAdd) {
		for (j=0; j<GetNumberOfVertices(); j++)
			vertices_tag[j] = 0;
	}

	for (i=0; i<vertex_tag_names.GetSize(); i++) {
		if (vertex_tag_names[i] == name) {
			// load
			par = (CUIntArray *)vertex_tag_arrays[i];
			for (j=0; j<par->GetSize(); j++)
				vertices_tag[par->GetAt(j)] = 1;
			return;
		}
	}
}

void CBINModel::GetFaceGroupNames(CStringArray &arr)
{
	arr.RemoveAll();
	
	int i;

	for (i=0; i<face_tag_names.GetSize(); i++) {
		arr.Add(face_tag_names[i]);
	}
}

void CBINModel::GetVertexGroupNames(CStringArray &arr)
{
	arr.RemoveAll();
	
	int i;

	for (i=0; i<vertex_tag_names.GetSize(); i++) {
		arr.Add(vertex_tag_names[i]);
	}
}

void CBINModel::FaceGroupDeleteFace(int num)
{
	// Delete deleted face from all groups,
	// adjust face indices accordingly

	CUIntArray *par;

	for (int i=0; i<face_tag_arrays.GetSize(); i++) {
		par = (CUIntArray *)face_tag_arrays[i];

		for (int j=0; j<par->GetSize(); j++) {
			if (par->GetAt(j)==num) {
				par->RemoveAt(j);
				break;
			}
		}

		for (j=0; j<par->GetSize(); j++) {
			if (par->GetAt(j)>num) {
				par->SetAt(j, par->GetAt(j)-1);
			}
		}
	}
}

void CBINModel::FaceGroupsSerialize(CStdioFile &file, CWordArray &cwaTrans)
{
	CUIntArray *par;
	char buf[500];

	file.WriteString(itoa(face_tag_arrays.GetSize(), buf, 10)+CString("\n"));

	for (int i=0; i<face_tag_arrays.GetSize(); i++) {
	
		file.WriteString(face_tag_names[i]+CString("\n"));
		
		par = (CUIntArray *)face_tag_arrays[i];

		file.WriteString(itoa(par->GetSize(), buf, 10)+CString("\n"));
		
		for (int j=0; j<par->GetSize(); j++) {
			file.WriteString(itoa(cwaTrans[par->GetAt(j)], buf, 10)+CString("\n"));
		}

	}
}

void CBINModel::VertexGroupsSerialize(CStdioFile &file)
{
	CUIntArray *par;
	char buf[500];

	file.WriteString(itoa(vertex_tag_arrays.GetSize(), buf, 10)+CString("\n"));

	for (int i=0; i<vertex_tag_arrays.GetSize(); i++) {
	
		file.WriteString(vertex_tag_names[i]+CString("\n"));
		
		par = (CUIntArray *)vertex_tag_arrays[i];

		file.WriteString(itoa(par->GetSize(), buf, 10)+CString("\n"));
		
		for (int j=0; j<par->GetSize(); j++) {
			file.WriteString(itoa(par->GetAt(j), buf, 10)+CString("\n"));
		}

	}
}

void CBINModel::FaceGroupsRead(CStdioFile &file, int iOldNum)
{
	CUIntArray *par;
	CString aux;
	CString csOldName;
	int num;
	char buf[5];

	int groupnum, facenum;

	file.ReadString(aux);
	groupnum = atoi(aux);

	for (int i=0; i<groupnum; i++) {

		file.ReadString(aux);
		
		csOldName = aux;
		num = 1;
		while (CopTools::CStringInCStringArray(aux, face_tag_names)) {
			aux = csOldName + "[" + CString(itoa(num, buf, 10)) +"]";
			num++;
		}
			
		face_tag_names.Add(aux);

		file.ReadString(aux);
		facenum = atoi(aux);
		
		par = new CUIntArray;

		for (int j=0; j<facenum; j++) {
			file.ReadString(aux);
			par->Add(atoi(aux)+iOldNum);
		}
		face_tag_arrays.Add(par);
	}
}

void CBINModel::VertexGroupsRead(CStdioFile &file, int iOldNum)
{
	CUIntArray *par;
	CString aux;
	CString csOldName;
	char buf[5];
	int num;

	int groupnum, vertexnum;

	file.ReadString(aux);
	groupnum = atoi(aux);

	for (int i=0; i<groupnum; i++) {

		file.ReadString(aux);

		csOldName = aux;
		num = 1;
		while (CopTools::CStringInCStringArray(aux, vertex_tag_names)) {
			aux = csOldName + "[" + CString(itoa(num, buf, 10)) +"]";
			num++;
		}
			
		vertex_tag_names.Add(aux);

		file.ReadString(aux);
		vertexnum = atoi(aux);
		
		par = new CUIntArray;

		for (int j=0; j<vertexnum; j++) {
			file.ReadString(aux);
			par->Add(atoi(aux)+iOldNum);
		}
		vertex_tag_arrays.Add(par);
	}
}

int CBINModel::ImportWavefront(CString csFile, int iFaceType)
{
	int old_faces = GetNumberOfFaces();
	int old_verts = GetNumberOfVertices();
	int i;
	
	// Open the Wavefront file
	CStdioFile imp;

	imp.Open(csFile, CFile::modeRead, NULL);

	CString aux = "";
	double x, y, z;
	int a, b, c, d;

	while (imp.ReadString(aux)) {
		aux.MakeUpper();
		if (aux[0] == 'V') {
			//TRACE("VERTEX: %s\n", aux);
			z = atof(aux.Right(aux.GetLength()-aux.ReverseFind(' ')));
			aux = aux.Left(aux.ReverseFind(' '));
			y = atof(aux.Right(aux.GetLength()-aux.ReverseFind(' ')));
			aux = aux.Left(aux.ReverseFind(' '));
			x = atof(aux.Right(aux.GetLength()-aux.ReverseFind(' ')));
			//TRACE("  %f, %f, %f\n", x,y,z);
			AddVertex(x*10000, y*10000, z*10000);
		}
		else if (aux[0] == 'S') {
			TRACE("S ON/OFF: %s\n", aux);
		}
		else if (aux[0] == 'G') {
			TRACE("G: %s\n", aux);
		}
		else if (aux[0] == 'U') {
			TRACE("Material: %s\n", aux);
		}
		else if (aux[0] == 'F') {
			//TRACE("FACE: %s\n", aux);
			d = atoi(aux.Right(aux.GetLength()-aux.ReverseFind(' ')));
			aux = aux.Left(aux.ReverseFind(' '));
			c = atoi(aux.Right(aux.GetLength()-aux.ReverseFind(' ')));
			aux = aux.Left(aux.ReverseFind(' '));
			b = atoi(aux.Right(aux.GetLength()-aux.ReverseFind(' ')));
			aux = aux.Left(aux.ReverseFind(' '));
			a = atoi(aux.Right(aux.GetLength()-aux.ReverseFind(' ')));
			a--; b--; c--; d--;
			if (a == -1) {
				a = b; b = c; c = d; d = -1;
			}
			//TRACE("  %d, %d, %d, %d\n", a,b,c,d);
			AddFace(a,b,c,d, iFaceType);
		}
	}
	imp.Close();
	return 1;
}

int CBINModel::GetNumberOfFaces()
{
	return faces.GetSize();
}

void CBINModel::WriteTexturesToStdioFile(CStdioFile &csfFile)
{
	for (int i=0; i<GetNumberOfTextures(); i++) {
		if (IsTextureInUse(csaBINTextures[i]))
			csfFile.WriteString(csaBINTextures[i]+"\n");
	}
}

void CBINModel::ExportWavefront(CString csFilename)
{
	char buf[100];
	CStdioFile exp;

	exp.Open(csFilename, CFile::modeCreate + CFile::modeWrite, NULL);

	for (int i=0; i<GetNumberOfVertices(); i++) {
		exp.WriteString("v "+CString(itoa(vertices[3*i+0], buf, 10))+" "+
							 CString(itoa(vertices[3*i+1], buf, 10))+" "+
							 CString(itoa(vertices[3*i+2], buf, 10))+"\n");
	}

	face *curface;
	for (i=0; i<GetNumberOfFaces(); i++) {
		curface = (face *)faces[i];

		exp.WriteString("f");
		for (int j=0; j<curface->points.GetSize(); j++) {
			exp.WriteString(" "+CString(itoa(curface->points[j]+1, buf, 10)));
		}
		exp.WriteString("\n");
	}

	exp.Close();
}

void CBINModel::FaceTextureFlipX(int iFace)
{
	if (iFace == -1) return;

	face *faceCurrent = (face *)faces[iFace];

	UINT uiDummy;

	uiDummy = faceCurrent->tex_x[0];
	faceCurrent->tex_x[0] = faceCurrent->tex_x[1];
	faceCurrent->tex_x[1] = uiDummy;

	if (faceCurrent->points.GetSize() != 4) return;
	
	uiDummy = faceCurrent->tex_x[2];
	faceCurrent->tex_x[2] = faceCurrent->tex_x[3];
	faceCurrent->tex_x[3] = uiDummy;
}

void CBINModel::FaceTextureFlipY(int iFace)
{
	if (iFace == -1) return;

	face *faceCurrent = (face *)faces[iFace];

	UINT uiDummy;

	uiDummy = faceCurrent->tex_y[0];
	faceCurrent->tex_y[0] = faceCurrent->tex_y[1];
	faceCurrent->tex_y[1] = uiDummy;

	if (faceCurrent->points.GetSize() != 4) return;
	
	uiDummy = faceCurrent->tex_y[2];
	faceCurrent->tex_y[2] = faceCurrent->tex_y[3];
	faceCurrent->tex_y[3] = uiDummy;
}

void CBINModel::FaceTextureRotate(int iFace)
{
	if (iFace == -1) return;

	face *faceCurrent = (face *)faces[iFace];

	UINT uiDummy;

	if (faceCurrent->points.GetSize() == 4) {
		
		uiDummy = faceCurrent->tex_x[0];
		faceCurrent->tex_x[0] = faceCurrent->tex_x[1];
		faceCurrent->tex_x[1] = faceCurrent->tex_x[2];
		faceCurrent->tex_x[2] = faceCurrent->tex_x[3];
		faceCurrent->tex_x[3] = uiDummy;
		uiDummy = faceCurrent->tex_y[0];
		faceCurrent->tex_y[0] = faceCurrent->tex_y[1];
		faceCurrent->tex_y[1] = faceCurrent->tex_y[2];
		faceCurrent->tex_y[2] = faceCurrent->tex_y[3];
		faceCurrent->tex_y[3] = uiDummy;
	}

	else if (faceCurrent->points.GetSize() == 3) {
		
		uiDummy = faceCurrent->tex_x[0];
		faceCurrent->tex_x[0] = faceCurrent->tex_x[1];
		faceCurrent->tex_x[1] = faceCurrent->tex_x[2];
		faceCurrent->tex_x[2] = uiDummy;
		uiDummy = faceCurrent->tex_y[0];
		faceCurrent->tex_y[0] = faceCurrent->tex_y[1];
		faceCurrent->tex_y[1] = faceCurrent->tex_y[2];
		faceCurrent->tex_y[2] = uiDummy;
	}
}

// Delete Vertex with the zero based number iVertex
// result:	1								- Success
//			DELETEVERTEX_VERTEXINFACE = -1	- Vertex in use by a face.
int CBINModel::DeleteVertex(int iVertex)
{
	// 1. check validity of index iVertex
	if (iVertex<0 || iVertex>GetNumberOfVertices()) return 0;

	// 2. check if vertex is in use
	if (VertexInUse(iVertex)) return DELETEVERTEX_VERTEXINFACE;

	// 3. remove it from the list of vertices
	vertices.RemoveAt(3*iVertex+2);		// z-coordinate
	vertices.RemoveAt(3*iVertex+1);		// y
	vertices.RemoveAt(3*iVertex);		// x

	// 4. adjust the vertex numbers in the faces accordingly
	face *pfFace;

	for (int i=0; i<GetNumberOfFaces(); i++) {
		pfFace = (face *)faces[i];
		for (int j=0; j<pfFace->points.GetSize(); j++)
			if (pfFace->points[j] > iVertex) pfFace->points[j]--;
	}

	// 5. adjust the vertex numbers in the vertex groups
	VertexGroupsDeleteVertex(iVertex);

	vertices_tag.SetSize(GetNumberOfVertices());

	return 1;
}

// Check if vertex is used by any face
BOOL CBINModel::VertexInUse(int iVertex)
{
	face *pfFace;

	for (int i=0; i<GetNumberOfFaces(); i++) {
		pfFace = (face *)faces[i];
		for (int j=0; j<pfFace->points.GetSize(); j++)
			if (pfFace->points[j] == iVertex) return TRUE;
	}

	return FALSE;
}

int CBINModel::VertexGroupsDeleteVertex(int iVertex)
{
	// Delete deleted vertex from all groups,
	// adjust vertex indices accordingly

	CUIntArray *cuaTaglist;

	for (int i=0; i<vertex_tag_arrays.GetSize(); i++) {
		cuaTaglist = (CUIntArray *)vertex_tag_arrays[i];

		for (int j=0; j<cuaTaglist->GetSize(); j++) {
			if (cuaTaglist->GetAt(j)==iVertex) {
				cuaTaglist->RemoveAt(j);
				break;
			}
		}

		for (j=0; j<cuaTaglist->GetSize(); j++) {
			if (cuaTaglist->GetAt(j)>iVertex) {
				cuaTaglist->SetAt(j, cuaTaglist->GetAt(j)-1);
			}
		}
	}

	return 1;
}

int CBINModel::GetNumberOfVertices()
{
	return vertices.GetSize()/3;
}

int CBINModel::AddTexAnim(CString csName, DWORD dwDelay, CStringArray &csaTextures, CString csARTDir)
{
	sTexAnim *psAnim = new sTexAnim;

	TRACE("Name: %s\n", csName);
	psAnim->csName = csName;
	TRACE("Delay: %ld\n", dwDelay);
	psAnim->dwDelay = dwDelay;
	for (int i=0; i<csaTextures.GetSize(); i++) {
		TRACE("%s\n",csaTextures[i]);
		psAnim->csaTextures.Add(csaTextures[i]);
		LoadTexture(csaTextures[i], csARTDir);
	}
	cpaTexAnims.Add(psAnim);

	return 1;
}

CString CBINModel::GetFirstAnimTexture(CString csAnim)
{
	TRACE("CBINModel::GetFirstAnimTexture(%s)\n", csAnim);
	
	
	sTexAnim *psAux;

	for (int i=0; i<cpaTexAnims.GetSize(); i++) {
		psAux = (sTexAnim *)cpaTexAnims[i];
		if (psAux->csName == csAnim) {
			TRACE("CBINModel::GetFirstAnimTexture = %s\n", psAux->csaTextures[0]);
			return psAux->csaTextures[0];
		}
	}

	return "";
}

int CBINModel::GetAnimNumFrames(CString csAnim)
{
	sTexAnim *psAux;

	for (int i=0; i<cpaTexAnims.GetSize(); i++) {
		psAux = (sTexAnim *)cpaTexAnims[i];
		if (psAux->csName == csAnim) {
			return psAux->csaTextures.GetSize();
		}
	}

	return 0;
}

CString CBINModel::GetAnimFrame(CString csAnim, int iFrame)
{
	sTexAnim *psAux;

	for (int i=0; i<cpaTexAnims.GetSize(); i++) {
		psAux = (sTexAnim *)cpaTexAnims[i];
		if (psAux->csName == csAnim) {
			return psAux->csaTextures[iFrame];
		}
	}

	return "";
}

DWORD CBINModel::GetAnimDelay(CString csAnim)
{
	sTexAnim *psAux;

	for (int i=0; i<cpaTexAnims.GetSize(); i++) {
		psAux = (sTexAnim *)cpaTexAnims[i];
		if (psAux->csName == csAnim) {
			return psAux->dwDelay;
		}
	}

	return 0;
}

void CBINModel::NewBIN(void)
{
	// clean Vertice Array
	vertices.RemoveAll();
	current_vertex = -1;
	vertices_tag.RemoveAll();

	// clean Face Array
	face *faux;
	for (int i=0; i<GetNumberOfFaces(); i++) {
		faux = (face *)faces[i];
		faux->points.RemoveAll();
		faux->tex_x.RemoveAll();
		faux->tex_y.RemoveAll();
		faux->texture.Empty();
		delete faces[i];
	}
	faces.RemoveAll();

	current_face = -1;

	// tags and groups
	CUIntArray *pAux;
	face_tag_names.RemoveAll();
	for (i=0; i<face_tag_arrays.GetSize(); i++) {
		pAux =	(CUIntArray *)face_tag_arrays[i];
		pAux->RemoveAll();
		delete pAux;
	}
	face_tag_arrays.RemoveAll();
	vertex_tag_names.RemoveAll();
	for (i=0; i<vertex_tag_arrays.GetSize(); i++) {
		pAux =	(CUIntArray *)vertex_tag_arrays[i];
		pAux->RemoveAll();
		delete pAux;
	}
	vertex_tag_arrays.RemoveAll();

	// Clean current tag arrays
	for (i=0; i<faces_tag.GetSize(); i++)
		faces_tag[i] = 0;
	for (i=0; i<vertices_tag.GetSize(); i++)
		vertices_tag[i] = 0;

	current_face = -1;
	liHellMagic = 0x00010000;	// MTM Size
}

BOOL CBINModel::IsVertexTagged(int iVertex)
{
	if (iVertex<0 || iVertex>GetNumberOfVertices()) return 0;
	return vertices_tag[iVertex];
}

int CBINModel::GetFaceVertex(int iFace, int iVertex)
{
	face *pFace = (face *)faces[iFace];

	if (iVertex >= pFace->points.GetSize()) return -1;

	return pFace->points[iVertex];
}

int CBINModel::GetFaceTextureType(int iFace)
{
	face *pFace = (face *)faces[iFace];

	return pFace->type;
}

void CBINModel::SetFaceVertex(int iFace, int iIndex, int iVertex)
{
	face *pFace = (face *)faces[iFace];

	// change size, if necessary
	if ((iIndex == 3) && (iVertex == -1) && (pFace->points.GetSize() == 4)) pFace->points.SetSize(3);
	if ((iIndex == 3) && (iVertex != -1) && (pFace->points.GetSize() == 3)) pFace->points.SetSize(4);

	if (iVertex != -1) pFace->points[iIndex] = iVertex;
}

void CBINModel::SetFaceTextureType(int iFace, int iTexType)
{
	face *pFace = (face *)faces[iFace];

	pFace->type = iTexType;
}

int CBINModel::Init(CString csDefaultACT)
{
	TRACE("CBINModel::Init(%s)\n", csDefaultACT);

	current_face	= -1;	// no face selected
	current_vertex	= -1;	// no vertex selected
	zoom			= 1.0;	// no zoom
	wAnimCnt		= 0;	// no RAW animation defined
	
	csDefPal = csDefaultACT;

	// Load default palette, usually [mtmPath]\ART\METALCR2.ACT
	CFile pal;

	if (!pal.Open(csDefPal, CFile::modeRead, NULL)) 
		AfxMessageBox("Error opening Default Palette file.");
		
	if (pal.Read(palette, 3*256)<3*256)
		AfxMessageBox("Palette Too Small!\n");
	
	pal.Close();

	bIsInitialized = TRUE;
	
	return 1;
}

BOOL CBINModel::IsVertexInFace(int iVertex, int iFace)
{
	if (!ValidVertex(iVertex)) return FALSE;	
	if (!ValidFace(iFace)) return FALSE;

	for (int i=0; i<4; i++) {
		if (GetFaceVertex(iFace, i) == iVertex) return TRUE;
	}
	return FALSE;
}

BOOL CBINModel::ValidFace(int iFace)
{
	return (iFace>=0 && iFace<=faces.GetUpperBound());
}

BOOL CBINModel::ValidVertex(int iVertex)
{
	return (iVertex>=0 && iVertex<=GetNumberOfVertices());
}

void CBINModel::GetTextureNameArray(CStringArray &csaTexNames)
{
	csaTexNames.RemoveAll();

	// add normal texture names
	for (int i=0; i<csaBINTextures.GetSize(); i++)
		csaTexNames.Add(csaBINTextures[i]);

	// add texture animation names
	sTexAnim *ptaAux;
	for (i=0; i<cpaTexAnims.GetSize(); i++) {
		ptaAux = (sTexAnim *)cpaTexAnims[i];
		csaTexNames.Add("ANIM:"+ptaAux->csName);
	}
}

CString CBINModel::GetFaceTextureName(int iFace)
{
	if (!ValidFace(iFace)) return CString("");
	
	face *curface = (face *)faces[iFace];
	return curface->texture;
}

int CBINModel::GetFaceTextureSize(int iFace)
{
	if (!ValidFace(iFace)) return -1;
	
	face *curface = (face *)faces[iFace];
	return curface->tex_size;
}

void CBINModel::GetFaceGroupVertices(CWordArray &cwaVertices)
{
	CByteArray cbaIsTagged;
	CWordArray cwaFaces;

	int i, iAux;

	GetFaceGroupFaces(cwaFaces);

	cbaIsTagged.SetSize(GetNumberOfVertices());

	for(i=0; i<GetNumberOfVertices(); i++)
		cbaIsTagged[i] = 0;

	for(i=0; i<cwaFaces.GetSize(); i++) {
		cbaIsTagged[GetFaceVertex(cwaFaces[i], 0)] = 1;
		cbaIsTagged[GetFaceVertex(cwaFaces[i], 1)] = 1;
		cbaIsTagged[GetFaceVertex(cwaFaces[i], 2)] = 1;
		
		iAux = GetFaceVertex(cwaFaces[i], 3);
		if (ValidVertex(iAux)) cbaIsTagged[iAux] = 1;
	}

	for(i=0; i<GetNumberOfVertices(); i++) {
		if (cbaIsTagged[i]) cwaVertices.Add(i);
	}
}

BOOL CBINModel::IsFaceTagged(int iFace)
{
	if(!ValidFace(iFace)) return FALSE;

	return faces_tag[iFace];
}

void CBINModel::GetFaceGroupFaces(CWordArray & cwaFaces)
{
	cwaFaces.RemoveAll();

	for(int i=0; i<GetNumberOfFaces(); i++) {
		if (IsFaceTagged(i)) cwaFaces.Add(i);
	}
}

int CBINModel::GetFaceNumberOfVertices(int iFace)
{
	if (!ValidFace(iFace)) return 0;

	if (GetFaceVertex(iFace, 3) == -1) return 3;
	return 4;
}

void CBINModel::GetFaceVertexList(int iFace, CWordArray &cwaVerts)
{
	cwaVerts.RemoveAll();
	if (!ValidFace(iFace)) return;
	
	for (int i=0; i<GetFaceNumberOfVertices(iFace); i++)
		cwaVerts.Add(GetFaceVertex(iFace, i));
}

void CBINModel::SetFaceVertexTextureCoords(int iFace, int iVertex, double dU, double dV)
{
	if (!ValidFace(iFace)) return;

	int iSize = GetFaceTextureSize(iFace);

	//if (iSize == 64) 
	//	dU = 1.0 - dU;
	

	face *curface = (face *)faces[iFace];

	dU = ((double)dU * (double)(0xFF0000));
	dV = ((double)dV * (double)(0xFF0000));

	curface->tex_x[iVertex] = (int)dU;
	curface->tex_y[iVertex] = (int)dV;
}

void CBINModel::ReadInfoFile(CString csName, int iOldVertNum, int iOldFaceNum)
{
	if (CopTools::FileExists(csName+".info")) {
		CStdioFile groups;
		groups.Open(csName+".info", CFile::modeRead, NULL);
		FaceGroupsRead(groups, iOldFaceNum);
		VertexGroupsRead(groups, iOldVertNum);
		groups.Close();
	}
}

// Merge all Vertices up to a maximal distance
int CBINModel::MergeCloseVertices(int iMaxDist)
{
	int iNumVert = GetNumberOfVertices();
	int iMerged = 0;

	for (int i=0; i<iNumVert-1; i++) {
		for (int j=i+1; j<iNumVert; j++) {
			if (VertexDistance(i, j) <= iMaxDist) {
				for (int k=0; k<GetNumberOfFaces(); k++) {
					for (int l=0; l<GetFaceNumberOfVertices(k); l++) {
						if (GetFaceVertex(k, l) == j) {
							SetFaceVertex(k, l, i);
							iMerged++;
						}
					}
				}
			}
		}
	}

	return iMerged;
}

long int CBINModel::VertexDistance(int iVert1, int iVert2)
{
	if (!ValidVertex(iVert1) || !ValidVertex(iVert2)) return -1;
	int x1, y1, z1;
	int x2, y2, z2;

	GetVertexCoords(iVert1, x1, y1, z1);
	GetVertexCoords(iVert2, x2, y2, z2);

	x1 = x1-x2;
	y1 = y1-y2;
	z1 = z1-z2;
	return int(sqrt(x1*x1 + y1*y1 + z1*z1));
}

void CBINModel::WriteAsVRML(CString csName)
{
	char buf[500];
	int i, j, k, l, x, y, z;
	CStdioFile cfVRML;

	cfVRML.Open(csName, CFile::modeWrite + CFile::modeCreate);

	cfVRML.WriteString("#VRML V1.0 ascii\n");
	cfVRML.WriteString("#\n");
	cfVRML.WriteString("# VRML 1.0 object created by BINedit.\n");
	cfVRML.WriteString("# To get the latest version:\n");
	cfVRML.WriteString("# Visit http://www.viaregio.de/pieper and follow\n");
	cfVRML.WriteString("# the links to Oliver's MTM Web Site or contact\n");
	cfVRML.WriteString("# pieper@viaregio.de for details.\n");
	cfVRML.WriteString("#\n");

	cfVRML.WriteString("\nShapeHints { vertexOrdering  CLOCKWISE }\n");

	// Write Vertex Coordinates
	cfVRML.WriteString("DEF Koords Coordinate3 { point [\n");

	for (i=0; i<GetNumberOfVertices(); i++) {
		GetVertexCoords(i, x, y, z);
		cfVRML.WriteString(CString(_itoa(x, buf, 10))+" "+CString(_itoa(y, buf, 10))+" "+CString(_itoa(-z, buf, 10))+",\n");
	}

	cfVRML.WriteString("] }\n");

	CStringArray csaTex;
	GetTextureNameArray(csaTex);


	CWordArray cwaVerts;
	double u, v;
	CString csAux;

	for (i=0; i<csaTex.GetSize(); i++) {
		if (IsTextureInUse(csaTex[i])) {
			cfVRML.WriteString("Separator {\nUSE Koords\n");
			cfVRML.WriteString("Texture2 { filename \""+csaTex[i]+".JPG\" }\n");

			// first write all texture coordinates for the following faces
			cfVRML.WriteString("TextureCoordinate2 {\npoint [\n");

			for (j=0; j<GetNumberOfFaces(); j++) 
				if (GetFaceTextureName(j) == csaTex[i]) {
				GetFaceVertexList(j, cwaVerts);
			
				for (k=0; k<cwaVerts.GetSize(); k++) {
					GetFaceVertexTextureCoords(j, k, u, v);
					csAux.Format("%f %f,\n", u, 1.0-v);
					cfVRML.WriteString(csAux);
				}
			}
			cfVRML.WriteString("] }\n");

			// then write the face definitions

			cfVRML.WriteString("IndexedFaceSet {\ncoordIndex [\n");

			for (j=0; j<GetNumberOfFaces(); j++) 
				if (GetFaceTextureName(j) == csaTex[i]) {
				GetFaceVertexList(j, cwaVerts);
			
				for (k=0; k<cwaVerts.GetSize(); k++) {
					cfVRML.WriteString(CString(_itoa(cwaVerts[k], buf, 10))+",");
				}
				cfVRML.WriteString("-1,\n");
			}
			cfVRML.WriteString("]\n");

			// and add the texture coord indices	
			cfVRML.WriteString("textureCoordIndex [\n");

			int iCnt = 0;
			for (j=0; j<GetNumberOfFaces(); j++) 
				if (GetFaceTextureName(j) == csaTex[i]) {
				GetFaceVertexList(j, cwaVerts);
			
				for (k=0; k<cwaVerts.GetSize(); k++) {
					cfVRML.WriteString(CString(_itoa(iCnt, buf, 10))+",");
					iCnt++;
				}
				cfVRML.WriteString("-1,\n");
			}
			cfVRML.WriteString("] }\n");
			cfVRML.WriteString("}\n");
		}
	}

	cfVRML.Close();
}


/*void CBINModel::WriteAsVRML(CString csName)
{
	char buf[500];
	int i, x, y, z;
	CStdioFile cfVRML;

	cfVRML.Open(csName, CFile::modeWrite + CFile::modeCreate);

	cfVRML.WriteString("#VRML V1.0 ascii\n");


	// Write Vertex Coordinates
	cfVRML.WriteString("DEF Koords Coordinate3 { point [\n");

	for (i=0; i<GetNumberOfVertices(); i++) {
		GetVertexCoords(i, x, y, z);
		cfVRML.WriteString(CString(_itoa(x, buf, 10))+" "+CString(_itoa(y, buf, 10))+" "+CString(_itoa(z, buf, 10))+",\n");
	}

	cfVRML.WriteString("] }\n");

	cfVRML.WriteString("DEF Model Separator {\nUSE Koords\nIndexedFaceSet {\ncoordIndex [\n");

	CWordArray cwaVerts;
	for (i=0; i<GetNumberOfFaces(); i++) {
		GetFaceVertexList(i, cwaVerts);

		for (int j=0; j<cwaVerts.GetSize(); j++) {
			cfVRML.WriteString(CString(_itoa(cwaVerts[j], buf, 10))+",");
		}
		cfVRML.WriteString("-1,\n");
	}
	cfVRML.WriteString("] } }\n");

	cfVRML.Close();
}
*/

void CBINModel::GetFaceVertexTextureCoords(int iFace, int iVert, double & u, double & v)
{
	if (!ValidFace(iFace)) return;
	if (GetFaceVertex(iFace, iVert) == -1) return;

	face *curface = (face *)faces[iFace];

	u = (double)curface->tex_x[iVert]/(double)(0xFF0000);
	v = (double)curface->tex_y[iVert]/(double)(0xFF0000);
}


void CBINModel::ClearTexturePool()
{
	// clean tex_pointers, tex_names Array
	int i;

	void *pBuf;
	for (i=0; i<csaBINTextures.GetSize(); i++) {
		if (cmapTexPtrs.Lookup(csaBINTextures[i], pBuf)) free(pBuf);
		if (cmapBmpTexPtrs.Lookup(csaBINTextures[i], pBuf)) free(pBuf);
	}
	cmapTexPtrs.RemoveAll();
	cmapBmpTexPtrs.RemoveAll();
	cmapTexSizes.RemoveAll();

	// texture animations
	sTexAnim *psTemp;
	for (i=0; i<cpaTexAnims.GetSize(); i++) {
		psTemp = (sTexAnim *)cpaTexAnims.GetAt(i);
		psTemp->csName.Empty();
		psTemp->csaTextures.RemoveAll();
		delete psTemp;
	}
	cpaTexAnims.RemoveAll();
}

void CBINModel::TagFace(int iFace)
{
	if (!ValidFace(iFace)) return;
	faces_tag[iFace] = TRUE;
}

void CBINModel::UntagFace(int iFace)
{
	if (!ValidFace(iFace)) return;
	faces_tag[iFace] = FALSE;
}

void CBINModel::TagVertex(int iVertex)
{
	if (!ValidVertex(iVertex)) return;
	vertices_tag[iVertex] = TRUE;
}

void CBINModel::UntagVertex(int iVertex)
{
	if (!ValidVertex(iVertex)) return;
	vertices_tag[iVertex] = FALSE;
}

void CBINModel::InitFaceTypes()
{
	csaFaceTypeNames.RemoveAll();
	cuaFaceTypeValues.RemoveAll();

	cuaFaceTypeValues.Add(0x19);
	csaFaceTypeNames.Add("0x19 Ignore Texture Coordinates (MTM)");

	cuaFaceTypeValues.Add(0xE);
	csaFaceTypeNames.Add("0xE Hellbender Face Type");

	cuaFaceTypeValues.Add(0x18);
	csaFaceTypeNames.Add("0x18 Normal Texture (MTM)");

	cuaFaceTypeValues.Add(0x29);
	csaFaceTypeNames.Add("0x29 Shiny Texture (MTM2)");

	cuaFaceTypeValues.Add(0x11);
	csaFaceTypeNames.Add("0x11 Transparent Color #0 (MTM)");

	cuaFaceTypeValues.Add(0x33);
	csaFaceTypeNames.Add("0x33 Transparent Color #0 (MTM2)");

	cuaFaceTypeValues.Add(0x00);
	csaFaceTypeNames.Add("0x?? Unknown Face Type");
}

CString CBINModel::GetFaceTypeName(int iType)
{
	for (int i=0; i<cuaFaceTypeValues.GetSize(); i++) {
		if (cuaFaceTypeValues[i] == iType) return csaFaceTypeNames[i];
	}

	return csaFaceTypeNames[csaFaceTypeNames.GetUpperBound()];
}

int CBINModel::GetFaceTypeValue(CString csName)
{
	for (int i=0; i<csaFaceTypeNames.GetSize(); i++) {
		if (csaFaceTypeNames[i] == csName) return cuaFaceTypeValues[i];
	}

	return 0x18;	// if undefined, use standard texture type
}

void CBINModel::GetFaceTypeNameArray(CStringArray & csaAux)
{
	csaAux.RemoveAll();
	
	for (int i=0; i<csaFaceTypeNames.GetSize(); i++) {
		csaAux.Add(csaFaceTypeNames[i]);
	}
}

void CBINModel::FlipFace(int iFace)
{
	if (!ValidFace(iFace)) return;

	int iAux;
	double dU0, dV0;
	double dU1, dV1;
	double dU2, dV2;
	double dU3, dV3;

	if (GetFaceNumberOfVertices(iFace) == 3) {
	// in case of a triangle: invert order of vertices
		GetFaceVertexTextureCoords(iFace, 1, dU1, dV1);
		GetFaceVertexTextureCoords(iFace, 2, dU2, dV2);
		
		SetFaceVertexTextureCoords(iFace, 2, dU1, dV1);
		SetFaceVertexTextureCoords(iFace, 1, dU2, dV2);

		iAux = GetFaceVertex(iFace, 1);
		SetFaceVertex(iFace, 1, GetFaceVertex(iFace, 2));
		SetFaceVertex(iFace, 2, iAux);
	} else if (GetFaceNumberOfVertices(iFace) == 4) {
	// in case of a quadrangle: xchange 0<->1 and 2<->3
		GetFaceVertexTextureCoords(iFace, 0, dU0, dV0);
		GetFaceVertexTextureCoords(iFace, 1, dU1, dV1);
		GetFaceVertexTextureCoords(iFace, 2, dU2, dV2);
		GetFaceVertexTextureCoords(iFace, 3, dU3, dV3);

		SetFaceVertexTextureCoords(iFace, 1, dU0, dV0);
		SetFaceVertexTextureCoords(iFace, 0, dU1, dV1);
		SetFaceVertexTextureCoords(iFace, 3, dU2, dV2);
		SetFaceVertexTextureCoords(iFace, 2, dU3, dV3);
		
		iAux = GetFaceVertex(iFace, 0);
		SetFaceVertex(iFace, 0, GetFaceVertex(iFace, 1));
		SetFaceVertex(iFace, 1, iAux);

		iAux = GetFaceVertex(iFace, 2);
		SetFaceVertex(iFace, 2, GetFaceVertex(iFace, 3));
		SetFaceVertex(iFace, 3, iAux);
	}
}

void CBINModel::FaceGroupByDirection(CString csDir, double dTolerance)
{
	int i;
	long int nx, ny, nz;
	double dx, dy, dz;
	double dAngle;

	csDir.MakeUpper();

	for(i=0; i<GetNumberOfFaces(); i++) {

		GetFaceNormal(i, nx, ny, nz, true);
		dx = nx/65000.0;
		dy = ny/65000.0;
		dz = nz/65000.0;
	
		if(csDir=="X") {
			if ((dx > 0) && (acos(dx) <= dTolerance)) TagFace(i);
		}

		else if(csDir=="-X") {
			if ((dx < 0) && (acos(-dx) <= dTolerance)) TagFace(i);
		}
		
		else if(csDir=="Y") {
			if ((dy > 0) && (acos(dy) <= dTolerance)) TagFace(i);
		}
		
		else if(csDir=="-Y") {
			if ((dy < 0) && (acos(-dy) <= dTolerance)) TagFace(i);
		}
		
		else if(csDir=="Z") {
			if ((dz > 0) && (acos(dz) <= dTolerance)) TagFace(i);
		}
		
		else if(csDir=="-Z") {
			if ((dz < 0) && (acos(-dz) <= dTolerance)) TagFace(i);
		}
	}
}

bool CBINModel::IsFaceTypeInUse(int iType)
{
	for (int i=0; i<GetNumberOfFaces(); i++) {
		if (GetFaceTextureType(i) == iType) return true;
	}

	return false;
}

void CBINModel::ImportASCModel(CString csName)
{
	TRACE("CBINModel::ImportASCModel(%s)\n", csName);
	CStdioFile asc;
	CString csLine, csAux;
	int i, v1, v2, v3;
	float x, y, z;

	asc.Open("C:\\Windows\\Desktop\\q2viper.asc", CFile::modeRead);
	
	asc.ReadString(csLine);
	asc.ReadString(csLine);

	csAux = CopTools::After(csLine, "Vertices:");
	csAux = CopTools::Before(csAux, "Faces:");
	int iVerts = atoi(csAux);

	csAux = CopTools::After(csLine, "Faces:");
	int iFaces = atoi(csAux);
	TRACE("  Verts: %d, Faces: %d\n", iVerts, iFaces);

	asc.ReadString(csLine);

	for (i=0; i<iVerts; i++) {
		asc.ReadString(csLine);

		csAux = CopTools::After(csLine, "X:");
		csAux = CopTools::Before(csAux, "Y:");
		csAux.SetAt(csAux.Find(','), '.');
		x = atof(csAux);

		csAux = CopTools::After(csLine, "Y:");
		csAux = CopTools::Before(csAux, "Z:");
		csAux.SetAt(csAux.Find(','), '.');
		y = atof(csAux);

		csAux = CopTools::After(csLine, "Z:");
		csAux.SetAt(csAux.Find(','), '.');
		z = atof(csAux);
		TRACE("  Vertex %d: (%f, %f, %f)\n", i, x, y, z);
		AddVertex(1000*x, 1000*y, 1000*z);
	}

	for (i=0; i<iFaces; i++) {
		asc.ReadString(csLine);

		csAux = CopTools::After(csLine, "A:");
		csAux = CopTools::Before(csAux, "B:");
		v1 = atoi(csAux);

		csAux = CopTools::After(csLine, "B:");
		csAux = CopTools::Before(csAux, "C:");
		v2 = atoi(csAux);

		csAux = CopTools::After(csLine, "C:");
		v3 = atoi(csAux);

		TRACE("  Face %d: (%d, %d, %d)\n", i, v1, v3, v2);
		AddFace(v1, v2, v3, -1, 0x18);
		asc.ReadString(csLine);
	}

	asc.Close();
}

void CBINModel::ImportASCSkin(CString csName)
{
	TRACE("CBINModel::ImportASCSkin(%s)\n", csName);
	CStdioFile asc;
	CString csLine, csAux;
	CUIntArray cuaU, cuaV;
	int i, v1, v2, v3;
	float x, y, z;

	asc.Open("C:\\Windows\\Desktop\\t2viper.asc", CFile::modeRead);
	
	asc.ReadString(csLine);
	asc.ReadString(csLine);

	csAux = CopTools::After(csLine, "Vertices:");
	csAux = CopTools::Before(csAux, "Faces:");
	int iVerts = atoi(csAux);

	csAux = CopTools::After(csLine, "Faces:");
	int iFaces = atoi(csAux);
	TRACE("  Verts: %d, Faces: %d\n", iVerts, iFaces);
	cuaU.SetSize(iVerts);
	cuaV.SetSize(iVerts);

	asc.ReadString(csLine);

	for (i=0; i<iVerts; i++) {
		asc.ReadString(csLine);

		csAux = CopTools::After(csLine, "X:");
		csAux = CopTools::Before(csAux, "Y:");
		cuaU[i] = atoi(csAux);

		csAux = CopTools::After(csLine, "Y:");
		csAux = CopTools::Before(csAux, "Z:");
		cuaV[i] = atoi(csAux);

		TRACE("  TexCoord %d: (%d, %d)\n", i, cuaU[i], cuaV[i]);
	}

	for (i=0; i<iFaces; i++) {
		asc.ReadString(csLine);

		csAux = CopTools::After(csLine, "A:");
		csAux = CopTools::Before(csAux, "B:");
		v1 = atoi(csAux);

		csAux = CopTools::After(csLine, "B:");
		csAux = CopTools::Before(csAux, "C:");
		v2 = atoi(csAux);

		csAux = CopTools::After(csLine, "C:");
		v3 = atoi(csAux);

		TRACE("  FaceTexCoordIndices %d: (%d, %d, %d)\n", i, v1, v2, v3);
		SetFaceVertexTextureCoords(i, 0, cuaU[v1]/304.0,  cuaV[v1]/195.0);
		SetFaceVertexTextureCoords(i, 1, cuaU[v2]/304.0,  cuaV[v2]/195.0);
		SetFaceVertexTextureCoords(i, 2, cuaU[v3]/304.0,  cuaV[v3]/195.0);
		asc.ReadString(csLine);
	}

	asc.Close();
}

void CBINModel::RebuildDisplayList()
{
	bRebuildDisplayList = true;
}

void CBINModel::SetFaceMagic(int iFace, long int iMagic)
{
	if (!ValidFace(iFace)) return;
	
	face *curface = (face *)faces[iFace];
	curface->unknown = iMagic;
}

long CBINModel::GetFaceMagic(int iFace)
{
	if (!ValidFace(iFace)) return 0;

	face *curface = (face *)faces[iFace];
	return curface->unknown;
}
