#ifndef __BINMODEL_H__
#define __BINMODEL_H__

#include <afxcoll.h>
#include <afxtempl.h>

// defines for function results
#define DELETEVERTEX_VERTEXINFACE -1

typedef struct facetag {
	int	type;
	int tex_size;
	BYTE tex_color[3];
	long int normal[3];
	CString texture;
	CUIntArray points;
	CUIntArray tex_x;
	CUIntArray tex_y;
	long int unknown;
} face;


typedef struct sTexAnimTag {
	CString csName;
	CStringArray csaTextures;
	DWORD dwDelay;
} sTexAnim;

class CBINModel {
	public:
		long GetFaceMagic(int iFace);
		void SetFaceMagic(int iFace, long int iMagic);
		void RebuildDisplayList(void);
		void ImportASCModel(CString csName);
		void ImportASCSkin(CString csName);
		bool IsFaceTypeInUse(int iType);
		void FaceGroupByDirection(CString csDir, double dTolerance);
		void FlipFace(int iFace);
		void GetFaceTypeNameArray(CStringArray &csaAux);
		int GetFaceTypeValue(CString csName);
		CString GetFaceTypeName(int iType);
		void UntagVertex(int iVertex);
		void TagVertex(int iVertex);
		void UntagFace(int iFace);
		void TagFace(int iFace);
		void ClearTexturePool(void);
		void GetFaceVertexTextureCoords(int iFace, int iVert, double &u, double &v);
		void WriteAsVRML(CString csName);
		long int VertexDistance(int iVert1, int iVert2);
		int MergeCloseVertices(int iMaxDist);
		void ReadInfoFile(CString csName, int iOldVertNum, int iOldFaceNum);
		void SetFaceVertexTextureCoords(int iFace, int iVertex, double dU, double dV);
		void GetFaceVertexList(int iFace, CWordArray &cwaVerts);
		int GetFaceNumberOfVertices(int iFace);
		void GetFaceGroupFaces(CWordArray &cwaFaces);
		BOOL IsFaceTagged(int iFace);
		void GetFaceGroupVertices(CWordArray &cwaVertices);
		int GetFaceTextureSize(int iFace);
		CString GetFaceTextureName(int iFace);
	void GetTextureNameArray(CStringArray &csaTexNames);
	BOOL ValidVertex(int iVertex);
	BOOL ValidFace(int iFace);
	BOOL IsVertexInFace(int iVertex, int iFace);
	int Init(CString csDefaultACT);
	void SetFaceTextureType(int iFace, int iTexType);
	void SetFaceVertex(int iFace, int iIndex, int iVertex);
	int GetFaceTextureType(int iFace);
	int GetFaceVertex(int iFace, int iVertex);
	BOOL IsVertexTagged(int iVertex);
	void NewBIN(void);
	DWORD GetAnimDelay(CString csAnim);
	CString GetAnimFrame(CString csAnim, int iFrame);
	int GetAnimNumFrames(CString csAnim);
	CString GetFirstAnimTexture(CString csAnim);
	CPtrArray cpaTexAnims;
	int AddTexAnim(CString csName, DWORD dwDelay, CStringArray &csaTextures, CString csARTDir);
	int GetNumberOfVertices();
	int VertexGroupsDeleteVertex(int iVertex);
	BOOL VertexInUse(int iVertex);
	int DeleteVertex(int iVertex);
	void FaceTextureRotate(int iFace);
	void FaceTextureFlipY(int iFace);
	void FaceTextureFlipX(int iFace);
	void ExportWavefront(CString csFilename);
	void WriteTexturesToStdioFile(CStdioFile &csfFile);
	int GetNumberOfFaces();
	void SetFaceTexture(int iFace, CString csTexture);
	int ImportWavefront(CString csFile, int iFaceType);
	void FaceGroupsRead(CStdioFile &file, int iOldNum);
	void FaceGroupsSerialize(CStdioFile &file, CWordArray &cwaTrans);
	void VertexGroupsRead(CStdioFile &file, int iOldNum);
	void VertexGroupsSerialize(CStdioFile &file);
	void FaceGroupDeleteFace(int num);
	void GetFaceGroupNames(CStringArray &arr);
	void GetVertexGroupNames(CStringArray &arr);
	void SaveFaceGroup(CString name);
	void RestoreFaceGroup(CString name, BOOL bAdd);
	void SaveVertexGroup(CString name);
	void RestoreVertexGroup(CString name, BOOL bAdd);
	void ToggleFaceTag(int num);
	void ToggleVertexTag(int num);
	int GetNumberOfTextures();
	int Scale(double fx, double fy, double fz);
	int AddVertex(int nx, int ny, int nz);
	int GetVertexCoords(int vnum, int &x, int &y, int &z);
	int SetVertexCoords(int vnum, int x, int y, int z);
	int GetNumberOfUsedTextures();
	int DeleteFace(int fnum);
	int AddFace(int v1, int v2, int v3, int v4, int type);
	void Translate(int dx, int dy, int dz, BOOL bTaggedOnly);
	void RotateX(double deg);
	void RotateY(double deg);
	void RotateZ(double deg);
	int CalcFaceMagic(int fnum);
	CString GetInfoString();
	void SetCurrentVertexTextureCoords(double tx, double ty);
	void MoveCurrentVertexTextureCoords(int dx, int dy);
	void SetCurrentFaceTexture(CString name);
	int LoadTexture(CString name, CString csARTDir);
	CString SelectTexture();
	void DrawTextureCoordSquare(int x, int y, int w, int size, int iface, CDC *pDC);
	CString GetCurrentFaceTextureName();
	int GetCurrentFaceTextureSize();
	int GetTextureSize(CString name);
	BYTE * GetTexturePointer(CString name);
	BYTE * GetBmpTexturePointer(CString name);
	BOOL IsTextureInUse(CString texname);
	void CycleCurrentFaceVertex(BOOL reset);
	void GetFaceNormal(int num, long int &nx, long int &ny, long int &nz, BOOL normalize);
	void GetVertexNormal(int num, long int &nx, long int &ny, long int &nz);
	void TranslateVertex(int num, int dx, int dy, int dz, BOOL bTaggedOnly);
	void SetVertex(int num, int x, int y, int z);
	void GetVertex(int num, int &x, int &y, int &z);
	int GetMinCoord(CString xyz);
	int GetMaxCoord(CString xyz);
	void DrawSideViewX(CDC *pDC, int x0, int y0, int w, int h);
	void DrawSideViewY(CDC *pDC, int x0, int y0, int w, int h);
	void DrawSideViewZ(CDC *pDC, int x0, int y0, int w, int h);
	void SetZoom(double z);
	double GetZoom(void);
	BOOL SetCurrentVertex(int number);
	BOOL SetCurrentFace(int number);
	int GetCurrentVertex(void);
	int GetCurrentFace(void);
	int GetHellMagic(void);
	void WriteBIN(CString filename, BOOL bBackdrop, BOOL bVertNorm, CWordArray &cwaTrans, long int liHellMagic);
	void ASCIIDump(CString filename);
	void DrawBIN(int FaceType, BOOL bAxes, BOOL bVerts);
	int ReadBIN(CString filename, CString csARTDir);
	CBINModel();
	~CBINModel();
protected:
	long int liHellMagic;
	bool bRebuildDisplayList;
	void InitFaceTypes(void);
	CStringArray csaFaceTypeNames;
	CUIntArray cuaFaceTypeValues;
	BOOL bIsInitialized;
	CPtrArray face_tag_arrays;
	CPtrArray vertex_tag_arrays;
	CStringArray face_tag_names;
	CStringArray vertex_tag_names;
	CByteArray vertices_tag;
	CByteArray faces_tag;
	CUIntArray tex_sizes;
	int CrossProduct(double ax, double ay, double az, double bx, double by, double bz, double &cx, double &cy, double &cz);
	double zoom;
	int current_face;
	int current_vertex;
	WORD wAnimCnt;	// counter for animated textures
	CString name;
	CString csDefPal;
	CString csBINDir;
	CStringArray csaBINTextures;
	CMapStringToPtr cmapTexPtrs;
	CMapStringToPtr cmapBmpTexPtrs;
	CMap<CString, LPCSTR, int, int> cmapTexSizes;
	unsigned char palette[3*256];
	CPtrArray faces;
	CUIntArray vertices;
};

#endif
