// Class CopTools
// (c)opyright 1997, Oliver Pieper
// pieper@viaregio.de

#include "stdafx.h"
#include "opTools.h"

///////////////////////////////////////
// Constructor
CopTools::CopTools() {
}

///////////////////////////////////////
// Destructor, clean everything up
CopTools::~CopTools() {
}


///////////////////////////////////////////////////////
// Get number of selected files from multiple selection
int CopTools::FileMultSelNumber(LPSTR lpstrSel)
{
	if (lpstrSel == NULL) return 0;
	if (lpstrSel[0] == 0) return 0;

	int num = 0;
	while (lpstrSel[0] != 0) {
		num++;
		do lpstrSel++; while (lpstrSel[0] != 0);
		lpstrSel++;
	}
	
	// more than one selection: first string is directory name
	if (num > 1) num--;
	
	return num;
}

///////////////////////////////////////////////
// Get directory of multiple selection
CString CopTools::FileMultSelDir(LPSTR lpstrSel)
{
	if (FileMultSelNumber(lpstrSel) == 0) return CString("");

	if (FileMultSelNumber(lpstrSel) == 1) {
		CString path(lpstrSel);

		return path.Left(path.ReverseFind('\\'));
	}
	else {
		return CString(lpstrSel);
	}
	return CString("");
}

void CopTools::FileMultSelFiles(LPSTR lpstrSel, CStringArray *pStrArr)
{
	pStrArr->RemoveAll();

	if (FileMultSelNumber(lpstrSel) == 0) return;

	if (FileMultSelNumber(lpstrSel) == 1) {
		CString path(lpstrSel);

		pStrArr->Add(path.Right(path.GetLength()-path.ReverseFind('\\')-1));

		return;
	}
	else {
		int num = 0;
		while (lpstrSel[0] != 0) {
			if (num > 0) pStrArr->Add(CString(lpstrSel));
			num++;
			do lpstrSel++; while (lpstrSel[0] != 0);
			lpstrSel++;
		}

		return;
	}
	return;
}

BOOL CopTools::FileExists(CString csName)
{
	return (GetFileAttributes(csName) != 0xFFFFFFFF);
}

// Return string portion before search string
// example: Before("Hello, World out there", ", World") = "Hello"
CString CopTools::Before(CString csStr, CString csSearch)
{
	return csStr.Left(csStr.Find(csSearch));
}

// Return string portion before and including search string
// example: BeforeIncluding("Hello, World out there", ", World") = "Hello, World"
CString CopTools::BeforeIncluding(CString csStr, CString csSearch)
{
	return csStr.Left(csStr.Find(csSearch)+csSearch.GetLength());
}

// Return string portion after search string
// example: After("Hello, World out there", ", World") = " out there"
CString CopTools::After(CString csStr, CString csSearch)
{
	return csStr.Right(csStr.GetLength()-csStr.Find(csSearch)-csSearch.GetLength());
}

// Return string portion after and including search string
// example: AfterIncluding("Hello, World out there", ", World") = ", World out there"
CString CopTools::AfterIncluding(CString csStr, CString csSearch)
{
	return csStr.Right(csStr.GetLength()-csStr.Find(csSearch));
}

// 980303op
// Read all entries of a .CFG file into the CMap object.
// Entries must be of the form "KEYNAME Value" where KEYNAME is a word w/o blanks.
// Examples of entries:
//    XCOORD 100
//    YCOORD 211
//    Title This is the whole title
void CopTools::ReadCfgFile(CString csName, CMapStringToString & cmapCfg)
{
	cmapCfg.RemoveAll();

	if (FileExists(csName)) {
		CStdioFile csfCfg;
		csfCfg.Open(csName, CFile::modeRead, NULL);

		CString csAux;
		while (csfCfg.ReadString(csAux)) {
			cmapCfg.SetAt(Before(csAux, " "), After(csAux, " "));
			TRACE("CopTools::ReadCfgFile(%s) : Set '%s' to '%s'\n", csName, Before(csAux, " "), After(csAux, " "));
		}
		csfCfg.Close();
	}
}

int CopTools::GetCfgInt(CString csKey, CMapStringToString &cmapCfg, int iDefault)
{
	CString csAux;

	csKey.MakeUpper();

	if (!cmapCfg.Lookup(csKey, csAux)) {
		char buf[20];
		cmapCfg.SetAt(csKey, CString(_itoa(iDefault, buf, 10)));
		return iDefault;
	}
	
	return atoi(csAux);
}

COLORREF CopTools::GetCfgCOLORREF(CString csKey, CMapStringToString &cmapCfg, COLORREF rgbDefault)
{
	CString csAux;

	csKey.MakeUpper();

	if (!cmapCfg.Lookup(csKey, csAux)) {
		char buf[20];
		cmapCfg.SetAt(csKey, CString(_ltoa(long int(rgbDefault), buf, 10)));
		return COLORREF(rgbDefault);
	}
	
	return COLORREF(atol(csAux));
}

CString CopTools::GetCfgStr(CString csKey, CMapStringToString &cmapCfg, CString csDefault)
{
	CString csAux;

	csKey.MakeUpper();

	if (!cmapCfg.Lookup(csKey, csAux)) {
		cmapCfg.SetAt(csKey, csDefault);
		return csDefault;
	}

	return csAux;
}

// 980303op
void CopTools::WriteCfgFile(CString csName, CMapStringToString & cmapCfg)
{
	CStdioFile csfCfg;
	csfCfg.Open(csName, CFile::modeWrite | CFile::modeCreate, NULL);

	POSITION pos;
	CString csKey, csValue;

	// Iterate through the entire map
    for( pos = cmapCfg.GetStartPosition(); pos != NULL; )    {
		cmapCfg.GetNextAssoc(pos, csKey, csValue);
		TRACE("CopTools::WriteCfgFile(%s) : Key is '%s', value is '%s'\n", csName, csKey, csValue);

		csKey.MakeUpper();
		csfCfg.WriteString( csKey + " " + csValue + "\n" );
	}
	csfCfg.Close();
}


//// 980312op //////////////////////////////////////
// Add extension to file name if not already present
// Extension must be specified WITHOUT the "."
CString CopTools::EnsureFileExtension(CString csFile, CString csExt)
{
	csFile.TrimLeft();
	csFile.TrimRight();
	csExt.TrimLeft();
	csExt.TrimRight();
	csExt = "."+csExt;

	CString csUFile = csFile;
	CString csUExt = csExt;
	
	csUFile.MakeUpper();
	csUExt.MakeUpper();

	if (csUFile.Right(4) != csUExt)
		csFile += csExt;

	return csFile;
}

BOOL CopTools::CStringInCStringArray(CString csSearch, CStringArray & csaArray)
{
	for (int i = 0; i<csaArray.GetSize(); i++) {
		if (csaArray[i] == csSearch) return TRUE;
	}

	return FALSE;
}

CString CopTools::GetFileFromPath(CString csPath)
{
	if (csPath.ReverseFind('\\') == -1) return csPath;
	return csPath.Right(csPath.GetLength()-csPath.ReverseFind('\\')-1);
}
