////////////////////////////////////////////////////////////////////////////////////////////////// Class CMTMPODFile
// Class for MTM POD Files
// [uses MICROSOFT FOUNDATION CLASSES]
////////////////////////////////////////////////////////////////////////////////////////////////// Class CMTMPODFile
// (c)opyright 1997, 1998 Oliver Pieper
// pieper@ViaRegio.de
// http://www.ViaRegio.de/pieper
////////////////////////////////////////////////////////////////////////////////////////////////
// 980320 - New function ValidDir() returns status of dir_read
// 980222 - ReleaseDirEntries() now as separate function
// 970000 - First Version
////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MTMPODFile.h"
#include "c:\dev\code\generic classes\opTools.h"

///////////////////////////////////////
// Constructor
CMTMPODFile::CMTMPODFile() {
	dir_read = FALSE;
	pMap = new CMapStringToPtr(500);
}

///////////////////////////////////////
// Destructor, clean everything up
CMTMPODFile::~CMTMPODFile() {

	// Clear all dir entries, if there are any
	ReleaseDirEntries();
}


//////////////////////////////////////
// Read the directory entries
int CMTMPODFile::ReadDirectory(CString csName)
{
	TRACE(">> CMTMPodFile::ReadDirectory(%s)\n", csName);
	int i;
	csPODName = csName;		// remember POD name for later use
	struct PODDirEntry *pde;
	
	// Clear all dir entries, if there are any
	ReleaseDirEntries();

	// Open the file
	CFile pod;
	if (!pod.Open(csName, CFile::modeRead, NULL)) {
		AfxMessageBox("CMTMPODFile::ReadDirectory:\nError opening file:\n"+csName);
		return 0;
	}

	// read number of direntries
	pod.Seek(0, CFile::begin);
	pod.Read(&num_files,4);
	
	dir.SetSize(num_files);	// resize dir array accordingly

	//////////////////////////////////////////////////////
	// Read all directory entries
	char buf[50];
	CString csAux;

	pod.Seek(0x54, CFile::begin);
	for (i=0; i<(int)num_files;i++) {
		pde = new PODDirEntry;
		pod.Read(buf, 40);
		csAux = pde->filename = Buf2Filename(buf);
		pde->palettename = Buf2Palettename(buf);
		pde->offset  = *((DWORD *)(&buf[36]));
		pde->dwLnth  = *((DWORD *)(&buf[32]));
		pde->dwIndex = i;	// Index in dir-Array, if found by map access
		TRACE("POD Dir Entry: %s, %s, %ld, %ld\n", pde->filename, pde->palettename, pde->offset, pde->dwLnth);
		dir[i] = pde;

		csAux.MakeUpper();
		pMap->SetAt(csAux, pde);	// store reference in map, for fast access by file name
	}
	pod_len = pod.GetLength();

	pod.Close();

	dir_read = TRUE;

	return 1;
}

//////////////////////////////////////////////////////////
// return file position of file in POD, or -1 if not found
long int CMTMPODFile::GetFilePosition(CString csName)
{
	if (!dir_read) return -2;

	void *pAux;
	csName.MakeUpper();

	if ( pMap->Lookup(csName, pAux) )
		return ((struct PODDirEntry *)pAux)->offset;

	return -1;
}

//////////////////////////////////////////////////////
// return file size of file in POD, or -1 if not found
long int CMTMPODFile::GetFileSize(CString csName)
{
	if (!dir_read) return -2;

	csName.MakeUpper();
	void *pAux;
	if (!pMap->Lookup(csName, pAux)) return -1;

	struct PODDirEntry *pde = (struct PODDirEntry *)pAux;
	
	int i = pde->dwIndex;

	if (i<dir.GetUpperBound())
		return ((PODDirEntry *)dir[i+1])->offset - pde->offset;
	else if (i<=dir.GetUpperBound())
		return pod_len - pde->offset;
	
	return -1;
}

//////////////////////////////////////////////////////
// return file size of file in POD, or -1 if not found
long int CMTMPODFile::GetFileSize(WORD idx)
{
	if (!dir_read) return -2;

	if (idx<dir.GetUpperBound())
		return ((PODDirEntry *)dir[idx+1])->offset - ((PODDirEntry *)dir[idx])->offset;
	else if (idx<=dir.GetUpperBound())
		return pod_len - ((PODDirEntry *)dir[idx])->offset;
	else return -1;
}

////////////////////////////////////////////////
// extract file name from direntry buffer
CString CMTMPODFile::Buf2Filename(char *buf)
{
	return CString(buf);
}

////////////////////////////////////////////////
// extract palette name from direntry buffer
CString CMTMPODFile::Buf2Palettename(char *buf)
{
	// if filetype is not "RAW", it does not use a palette, return "<none>"
	CString aux = CString(buf).Right(3);
	aux.MakeUpper();
	if (aux!="RAW") return CString("<none>");

	int i=0;
	while (buf[i++]);

	aux = CString(&buf[i]);
	if (aux.IsEmpty()) aux = "<none>";
	return aux;
}

//////////////////////////////////////////////////
// Allocate memory for a file and load it
BYTE * CMTMPODFile::AllocMemAndLoadFile(CString csName)
{
	if (GetFilePosition(csName) == -1) return NULL;

	long int size = GetFileSize(csName);
	
	BYTE *buffer = (BYTE *)CopTools::GlobalAllocPtr(GMEM_FIXED, size);

	CFile pod;
	pod.Open(csPODName, CFile::modeRead, NULL);
	pod.Read(buffer, size);
	pod.Close();

	return buffer;
}

////////////////////////////////
// Return number of files in POD
WORD CMTMPODFile::GetFileNum()
{
	return (WORD)num_files;
}

///////////////////////////////////////////////////////////////////
// Fill the variables with the data of the direntry at position idx
int CMTMPODFile::GetDirEntryAt(WORD idx, CString &csName, CString &csPal, DWORD &dwOffset, DWORD &dwLen, DWORD &dwLnth)
{
	TRACE(">> CMTMPodFile::GetDirEntryAt(%d,...)\n", idx);
	
	if (idx < 0 || idx > GetFileNum()-1) return 0;
	
	PODDirEntry *pPDE = (PODDirEntry *)dir[idx];

	csName   = pPDE->filename;
	csPal    = pPDE->palettename;
	dwOffset = pPDE->offset;
	dwLnth   = pPDE->dwLnth;
	dwLen    = GetFileSize(idx);
	return 1;
}

DWORD CMTMPODFile::GetLength()
{
	if (!dir_read) return 0;
	return pod_len;
}

///970501op/////////////////////////////////
// Get number of files in any POD. Do not
// change directory entries.
WORD CMTMPODFile::GetPODFileNum(CString csName)
{
	CFile pod;
	if (!pod.Open(csName, CFile::modeRead, NULL)) {
		AfxMessageBox("CMTMPODFile::GetPODFileNum:\nError opening file:\n"+csName);
		return 0;
	}

	// read number of direntries
	WORD wNumFiles;
	pod.Read(&wNumFiles,4);
	pod.Close();
	return wNumFiles;
}

DWORD CMTMPODFile::GetFileLnth(CString csName)
{
	if (!dir_read) return 0;

	void *pAux;
	csName.MakeUpper();

	if ( pMap->Lookup(csName, pAux) )
		return ((struct PODDirEntry *)pAux)->dwLnth;

	return 0;
}

//980222op///////////////////////////////////////
// Release all dir entries to clean up memory
void CMTMPODFile::ReleaseDirEntries(void)
{
	struct PODDirEntry *pde;

	for (int i=0; i<dir.GetSize(); i++)	{
		pde = ((PODDirEntry *)dir[i]);
		pde->filename.Empty();			// important! release the CString objects
		pde->palettename.Empty();		// important! release the CString objects
		delete dir[i];
	}
	
	dir.RemoveAll();

	pMap->RemoveAll();
}

bool CMTMPODFile::ValidDir()
{
	return dir_read ? 1 : 0;
}
