// Class CMTMTools
// (c)opyright 1997, Oliver Pieper
// pieper@viaregio.de

#include "stdafx.h"
#include "MTMTools.h"
#include "C:\dev\Code\MTM Classes\MTMPODFile.h"
#include "C:\Dev\Code\Generic Classes\opTools.h"

///////////////////////////////////////
// Constructor
CMTMTools::CMTMTools() {
}

///////////////////////////////////////
// Destructor, clean everything up
CMTMTools::~CMTMTools() {
}


//////////////////////////////////////////////////
// merge POD files. Exclusion not yet implemented
int CMTMTools::MergePODs(CStringArray &merge, CStringArray &system, CString merged_name, CProgressCtrl *pcpcProg, CString *pcsInfo, CDialog *pcdlgParent)
{
	int i, j;
	merged_name.MakeUpper();

	////////////////////////////////
	// Check validity of parameters
	//
	if (merge.GetSize() == 0) {
		AfxMessageBox("MergePOD ERROR: No PODs specified!");
		return 0;
	}
	
	if (merged_name.IsEmpty()) {
		AfxMessageBox("MergePOD ERROR: No destination file name specified!");
		return 0;
	}

	for (i=0; i<merge.GetSize(); i++)
		if (merge[i]==merged_name) {
			AfxMessageBox("MergePOD ERROR: Destination POD \n"+merged_name+"\nmay not be source!");
			return 0;
		}

	if (CopTools::FileExists(merged_name)) {
		if (AfxMessageBox("Destination file\n"+merged_name+"\nexists! Overwrite?", MB_YESNO) != IDYES)
			return 0;
	}


	if (pcdlgParent) pcdlgParent->BeginWaitCursor();

	////////////////////////////////////
	// validity of parameters checked,
	// proceed with merging
	CMTMPODFile pod;
	CPtrArray cpaMergeList;
	CPtrArray syslist;
	sMLE *mle;				// struct scpaMergeListEntry
	DWORD dwOldSize = 0;	// sum of old filesizes
	DWORD dwOldFiles  = 0;	// number of old files
	DWORD dwProcFiles = 0;	// number of files processed

	// exclude files (not implemented)
	for (i=0; i<system.GetSize(); i++) {
		if (pcsInfo) *pcsInfo = "PODMate: Scanning exclude PODs";
		if (pcdlgParent) pcdlgParent->UpdateData(FALSE);
		TRACE("Exclude (not implemented!) %d: %s\n", i, system[i]);
	}
	
	// for pcpcProgress indicator: get number of all files
	for (i=0; i<merge.GetSize(); i++) dwOldFiles += CMTMPODFile::GetPODFileNum(merge[i]);
	TRACE("Number of source files: %d\n", dwOldFiles);
	if (pcpcProg) pcpcProg->SetRange(0, dwOldFiles);

	
	for (i=0; i<merge.GetSize(); i++) {
		if (pcsInfo) *pcsInfo = "PODMate: Pass 1 - Reading source POD directories";
		if (pcdlgParent) pcdlgParent->UpdateData(FALSE);
		
		pod.ReadDirectory(merge[i]);
		TRACE("Merge %d: %s (%d files)\n", i, merge[i], pod.GetFileNum());

		dwOldSize  += pod.GetLength();
		
		///////////////////////////////////////
		// transfer direntries to cpaMergeList
		for (j=0; j<pod.GetFileNum(); j++) {
			
			dwProcFiles++; if (pcpcProg) pcpcProg->SetPos(dwProcFiles);
			
			mle = new sMLE;
			mle->wSrc = i;	// Offset of source POD in CStringArray merge
			pod.GetDirEntryAt(j, mle->csName, mle->csPalette, mle->dwSrcOffset, mle->dwLen, mle->dwLnth);

			if (!CMTMTools::MergePODs_IsAlreadyMerged(mle->csName, cpaMergeList))
				cpaMergeList.Add(mle);
			else delete mle;

		}
	}

	////////////////////////////////////////////
	// recalculate mergelist destination offsets
	
	// size of header: 0x54
	// size of directory: 40 * number of files
	if (pcsInfo) *pcsInfo = "PODMate: Recalculating file offsets";
	if (pcdlgParent) pcdlgParent->UpdateData(FALSE);

	DWORD dwCurOffset = 0x54 + 40 * cpaMergeList.GetSize();
	
	for (i=0; i<cpaMergeList.GetSize(); i++) {
		mle = (sMLE *)cpaMergeList[i];
		mle->dwDstOffset = dwCurOffset;
		dwCurOffset += mle->dwLen;
	}

	//////////////////////////////////////////////
	// Write destination POD header and directory
	if (pcsInfo) *pcsInfo = "PODMate: Writing destination POD directory";
	if (pcdlgParent) pcdlgParent->UpdateData(FALSE);
	CFile cfDest;
	char nullmem[200];
	struct DirEntry {
		char pchNamePal[32];
		DWORD lnth;
		DWORD offset;
	} deCurEntry;
	
	cfDest.Open(merged_name, CFile::modeWrite + CFile::modeCreate);
	
	ZeroMemory(nullmem, 200); cfDest.Write(nullmem, 200);
	
	dwProcFiles = cpaMergeList.GetSize();

	cfDest.Seek(0, CFile::begin);	// number of files 
	cfDest.Write(&dwProcFiles, 4);
	cfDest.Seek(4, CFile::begin);
	char titel[] = "Merged by PODMate2.02, (c)1997 Oliver Pieper";
	cfDest.Write(titel, strlen(titel));
	cfDest.Seek(0x54, CFile::begin);

	for (i=0; i<cpaMergeList.GetSize(); i++) {
		mle = (sMLE *)cpaMergeList[i];
		ZeroMemory(deCurEntry.pchNamePal, 36);
		deCurEntry.offset = mle->dwDstOffset;
		deCurEntry.lnth   = mle->dwLnth;
		memcpy(deCurEntry.pchNamePal, (LPCTSTR)mle->csName, mle->csName.GetLength());
		if (mle->csPalette != "<none>")
			memcpy((void *)((DWORD)deCurEntry.pchNamePal + mle->csName.GetLength() + 1), (LPCTSTR)mle->csPalette, mle->csPalette.GetLength());
		cfDest.Write(&deCurEntry, 40);
	}

	//////////////////////////////////////////////
	// Copy data to destination
	if (pcsInfo) *pcsInfo = "PODMate: Pass 2 - Copying data";
	if (pcdlgParent) pcdlgParent->UpdateData(FALSE);

	if (pcpcProg) pcpcProg->SetRange(0, cpaMergeList.GetSize());
	
	CFile src;
	void *buf;
	WORD wCurIdx = 0xFFFF;

	for (i=0; i<cpaMergeList.GetSize(); i++) {

		if (pcpcProg) pcpcProg->SetPos(i);

		mle = (sMLE *)cpaMergeList[i];
		
		if (mle->wSrc != wCurIdx) {	// new POD file
			if (wCurIdx != 0xFFFF) src.Close();
			src.Open(merge[mle->wSrc], CFile::modeRead, NULL);
			wCurIdx = mle->wSrc;
		}

		buf = CopTools::GlobalAllocPtr(GMEM_FIXED, mle->dwLen);
		src.Seek(mle->dwSrcOffset, CFile::begin);
		src.Read(buf, mle->dwLen);
		cfDest.Write(buf, mle->dwLen);
		CopTools::GlobalFreePtr(buf);
	}
	src.Close();

	DWORD dwDstLen = cfDest.GetLength();

	cfDest.Close();

	TRACE("Size of source PODs: %d\n", dwOldSize);
	
	///////////////////////////////////////////////
	// report
	CString report;
	report  = CString("The source PODs contained ")+CString(ltoa(dwOldFiles, nullmem, 10)+CString(" files,\n"));
	report += CString(ltoa(dwOldFiles-cpaMergeList.GetSize(), nullmem, 10))+CString(" of them were duplicates.\n\n");
	report += CString("The merging resulted in a size reduction of ")+CString(ltoa((dwOldSize-dwDstLen)/1024, nullmem, 10))+CString(" kBytes.\n");
	AfxMessageBox(report, MB_OK);

	if (pcsInfo) *pcsInfo = "USER: Select Source and Destination PODs";
	if (pcdlgParent) pcdlgParent->UpdateData(FALSE);
	
	////////////////////////////////
	// Cleanup phase (long live MtG)
	for (i=0; i<cpaMergeList.GetSize(); i++) {
		mle = (sMLE *)cpaMergeList[i];
		mle->csName.Empty();
		mle->csPalette.Empty();
		delete mle;
	}
	cpaMergeList.RemoveAll();
	
	if (pcdlgParent) pcdlgParent->EndWaitCursor();
	return 1;
}

CString CMTMTools::GetMTMPath()
{
	// Try to obtain Registry info
	CString mtmPath;
	DWORD dwSize = 0;

	HKEY hKey = NULL;
	if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Monster Truck Madness\\1.0", 0, KEY_READ, &hKey ) == ERROR_SUCCESS )
	{
		DWORD dwSize = 0;
		// see how long the data is....
		if ( RegQueryValueEx( hKey, "InstalledTo", NULL, NULL, NULL, &dwSize ) == ERROR_SUCCESS )
		{
			LONG lRet = RegQueryValueEx( hKey, "InstalledTo", NULL, NULL, (LPBYTE)mtmPath.GetBuffer( dwSize ), &dwSize );
			mtmPath.ReleaseBuffer();
		}
		RegCloseKey( hKey );
	} else {
		// AfxMessageBox("Could not find\nMonster Truck Madness Path\nin Registry", MB_OK + MB_ICONEXCLAMATION);
		mtmPath = ".";
	}

	return mtmPath;
}


///970501op///////////////////////////////////
// belongs to MergePODs()
// - determine if csName is already in the array
BOOL CMTMTools::MergePODs_IsAlreadyMerged(CString &csName, CPtrArray &cpaMerged)
{
	int i;
	sMLE *mle;

	for (i=0; i<cpaMerged.GetSize(); i++) {
		mle = (sMLE *)cpaMerged[i];
		if (mle->csName == csName) return TRUE;
	}

	return FALSE;
}
